import inspect
import os
import time
from typing import Any, List, Optional, Protocol, Tuple

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray
from scipy import stats
from so3g.proj import RangesMatrix

from sotodlib.core import AxisManager


def plot_glitch_stats(
    tod,
    flags="flags",
    glitches="glitches",
    N_bins=30,
    save_path="./",
    save_name="glitch_flag_stats.png",
    save_plot=True,
):
    """
    Function for plotting the glitch flags/cut statistics using the built in stats functions
    in the RangesMatrices class.
    Args:
    -----
    tod (AxisManager): Axis manager with glitch flags in it.
    flags (FlagManager): Name of the FlagManager that holds the glitch flags in tod.
    glitches (RangesMatrix): Name of the axis inside flags that holds the RangesMatrix with
                             the glitches in it.
    N_bins (int): Number of bins in the histogram.
    save_path (path): Path to directory where plot will be saved
    save_name (str): Name of file to save plot to in save_path
    save_plot (bool): Determines if plot is saved.

    Returns:
    --------
    fig (matplotlib.figure.Figure): Figure class
    ax (numpy.ndarray): Numpy array of matplotlib.axes._subplots.AxesSubplot class plot axes.
    frac_samp_glitches (ndarray, float): Array size 1 x N_dets with number of samples flagged
                                         per detector divided by the total number of samples
                                         converted to percentage.
    interval_glitches (ndarray, int): Array size 1 x N_dets with number flagged intervals per
                                      detectors.
    """
    fig, axes = plt.subplots(1, 2, figsize=(15, 5))
    ax = axes.flatten()
    frac_samp_glitches = (
        100 * np.asarray(tod[flags][glitches].get_stats()["samples"]) / tod.samps.count
    )
    glitchlog = np.log10(frac_samp_glitches[frac_samp_glitches > 0])
    binmin = int(np.floor(np.min(glitchlog)))
    binmax = int(np.ceil(np.max(glitchlog)))
    _ = ax[0].hist(
        frac_samp_glitches, bins=np.logspace(binmin, binmax, N_bins), label="_nolegend_"
    )
    medsamps = np.median(frac_samp_glitches)
    ax[0].axvline(medsamps, color="C1", ls=":", lw=2, label=f"Median: {medsamps:.2f}%")
    meansamps = np.mean(frac_samp_glitches)
    ax[0].axvline(meansamps, color="C2", ls=":", lw=2, label=f"Mean: {meansamps:.2f}%")
    modesamps = stats.mode(frac_samp_glitches, keepdims=True)
    ax[0].axvline(
        modesamps[0][0],
        color="C3",
        ls=":",
        lw=2,
        label=f"Mode: {modesamps[0][0]:.2f}%, Counts: {modesamps[1][0]}",
    )
    stdsamps = np.std(frac_samp_glitches)
    ax[0].axvspan(
        meansamps - stdsamps,
        meansamps + stdsamps,
        color="wheat",
        alpha=0.2,
        label=f"$\\sigma$: {stdsamps:.2f}%",
    )
    ax[0].legend()
    ax[0].set_xlim(10**binmin, 10**binmax)
    ax[0].set_xscale("log")
    ax[0].set_xlabel("Fraction of Samples Flagged\nper Detector [%]", fontsize=16)
    ax[0].set_ylabel("Counts", fontsize=16)
    ax[0].set_title(
        "Samples Flagged Stats\n$N_{\\mathrm{dets}}$ = "
        + f"{tod.dets.count}"
        + " and $N_{\\mathrm{samps}}$ = "
        + f"{tod.samps.count}",
        fontsize=18,
    )

    interval_glitches = np.asarray(tod[flags][glitches].get_stats()["intervals"])
    binlinmax = np.nanmax(interval_glitches)
    _ = ax[1].hist(interval_glitches, bins=np.linspace(0, binlinmax, N_bins))
    medints = np.median(interval_glitches)
    ax[1].axvline(
        medints, color="C1", ls=":", lw=2, label=f"Median: {medints:.2f} intervals"
    )
    meanints = np.mean(interval_glitches)
    ax[1].axvline(
        meanints, color="C2", ls=":", lw=2, label=f"Mean: {meanints:.2f} intervals"
    )
    modeints = stats.mode(interval_glitches, keepdims=True)
    ax[1].axvline(
        modeints[0][0],
        color="C3",
        ls=":",
        lw=2,
        label=f"Mode: {modeints[0][0]:.2f} intervals, Counts: {modeints[1][0]}",
    )
    stdints = np.std(interval_glitches)
    ax[1].axvspan(
        meanints - stdints,
        meanints + stdints,
        color="wheat",
        alpha=0.2,
        label=f"$\\sigma$: {stdints:.2f} intervals",
    )

    ax[1].legend()
    ax[1].set_xlabel("Number of Flag Intervals\nper Detector", fontsize=16)
    ax[1].set_ylabel("Counts", fontsize=16)
    ax[1].set_title(
        "Ranges Flag Manager Stats\n$N_{\\mathrm{dets}}$ with $\\geq$ 1 interval = "
        + f"{len(interval_glitches[interval_glitches > 0])}/{tod.dets.count}",
        fontsize=18,
    )
    plt.suptitle(
        f"Glitch Stats for Obs Timestamp: {tod.obs_info.timestamp}, dT = {np.ptp(tod.timestamps)/60:.2f} min",
        fontsize=18,
    )
    save_ts = str(int(time.time()))
    plt.subplots_adjust(top=0.75, bottom=0.2)
    if save_plot:
        plt.savefig(os.path.join(save_path, save_ts + "_" + save_name))
    return fig, ax, frac_samp_glitches, interval_glitches


class FlagFunc(Protocol):
    def __call__(
        self,
        aman: AxisManager,
        *,
        signal: Optional[NDArray[np.floating]] = ...,
        merge: bool = ...,
        name=...,
        overwrite: bool = ...,
    ):
        ...


def flag_sliding_window(
    aman: AxisManager,
    flag_func: FlagFunc,
    signal: Optional[NDArray[np.floating]] = None,
    window_size: int = 10000,
    overlap: int = 1000,
    **kwargs,
) -> Tuple[RangesMatrix, List[Any]]:
    """
    Run a flag through a sliding window.
    Currently written only for flags that run on signal data.

    Arguments:

        aman: The AxisManager to pass to flagging function.

        flag_func: The flagging function.
                   First argument should be 'aman' and should also contain kwargs:
                   'signal', 'merge', 'name', 'overwrite'.

        signal: Data to flag, if None aman.signal is used.

        window_size: Size of window to use.

        overlap: Overlap between adjacent windows.

        **kwargs: kwargs to pass to flag_func.

    Returns:

        flag_ranges: RangesMatrix of the flag generated by flag_func.

        other: Any other outputs generated by flag_func,
               This is a list where each element is the outputs for an iteration of a window.
    """
    if signal is None:
        signal = aman.signal
    if not isinstance(signal, np.ndarray):
        raise TypeError("Signal is not an array")

    merge_arg_names = ["merge", "overwrite", "name"]
    sig = inspect.signature(flag_func)
    merge_args = {ma: sig.parameters[ma].default for ma in merge_arg_names}
    merge_args.update(kwargs)
    merge_args = {ma: merge_args[ma] for ma in merge_arg_names}
    kwargs["merge"] = False

    other = []
    flag = np.zeros(signal.shape, dtype=bool)
    tot = flag.shape[-1]
    # Could easily be parallelized here
    # numba, jax, concurrent.futures, or joblib all plug in well
    for i in range(tot // (window_size - overlap)):
        start = i * (window_size - overlap)
        end = np.min((start + window_size, tot))
        out = flag_func(
            aman,
            signal=signal[..., start:end],
            **kwargs,
        )
        if isinstance(out, tuple):
            _flag = out[0]
            other.append(out[1:])
        else:
            _flag = out
            other.append(None)
        flag[..., start:end] += _flag.mask()
    flag_ranges = RangesMatrix.from_bitmask(flag)

    if merge_args["merge"]:
        _merge(aman, flag_ranges, merge_args["name"], merge_args["overwrite"])

    return flag_ranges, other


def _merge(tod: AxisManager, flag_ranges: RangesMatrix, name: str, overwrite: bool):
    if not isinstance(tod, AxisManager):
        print("TOD is not an AxisManager, not merging")
        return
    elif "dets" not in tod or "samps" not in tod:
        print("dets or samps axis not in TOD, not merging")
        return
    elif flag_ranges.shape != (tod.dets.count, tod.samps.count):
        print("Shape of flag does not match that of TOD, not merging")
        return
    if name in tod.flags._fields:
        if overwrite:
            tod.flags.move(name, None)
        else:
            print("Flag already exists and overwrite is False")
    if "flags" not in tod._fields:
        flags = AxisManager(tod.dets, tod.samps)
        tod.wrap("flags", flags)
    tod.flags.wrap(name, flag_ranges, [(0, "dets"), (1, "samps")])



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Plot a hardware model.
"""

import argparse

import astropy.units as u

from ..core import Hardware

from ..vis_hardware import plot_detectors, summary_text


def main():
    parser = argparse.ArgumentParser(
        description="This program reads a hardware model and plots the\
            detectors.  Note that you should pre-select detectors before\
            passing a hardware model to this function.  See so_hardware_trim.",
        usage="so_hardware_plot [options] (use --help for details)")

    parser.add_argument(
        "--hardware", required=True, default=None,
        help="Input hardware file"
    )

    parser.add_argument(
        "--out", required=False, default=None,
        help="Name of the output PDF file."
    )

    parser.add_argument(
        "--width", required=False, default=None,
        help="The width of the plot in degrees."
    )

    parser.add_argument(
        "--height", required=False, default=None,
        help="The height of the plot in degrees."
    )

    parser.add_argument(
        "--labels", required=False, default=False, action="store_true",
        help="Add pixel and polarization labels to the plot."
    )

    parser.add_argument(
        "--show_centers", required=False, default=False, action="store_true",
        help="Add labels with pixel center coordinates."
    )

    parser.add_argument(
        "--xieta", required=False, default=False, action="store_true",
        help="Plot in Xi / Eta coordinates."
    )

    parser.add_argument(
        "--lat_corotate",
        required=False,
        default=False,
        action="store_true",
        help="Rotate LAT receiver to maintain focalplane orientation",
    )

    parser.add_argument(
        "--lat_elevation_deg",
        required=False,
        default=60.0,
        type=float,
        help="Observing elevation of the LAT if not co-rotating",
    )

    args = parser.parse_args()

    outfile = args.out
    if outfile is None:
        fields = args.hardware.split(".")
        outfile = fields[0]

    print("Loading hardware file {}...".format(args.hardware), flush=True)
    hw = Hardware(args.hardware)
    # summary_text(hw)

    width = args.width
    if width is not None:
        width = float(width)
    height = args.height
    if height is not None:
        height = float(height)

    print("Generating detector plot...", flush=True)
    plot_detectors(
        hw.data["detectors"],
        outfile,
        width=width,
        height=height,
        labels=args.labels,
        xieta=args.xieta,
        lat_corotate=args.lat_corotate,
        lat_elevation=args.lat_elevation_deg * u.degree,
        show_centers=args.show_centers
    )

    return



import argparse as ap
import datetime as dt
import logging
import os
from copy import deepcopy
from importlib import import_module
from typing import List, Optional

import git
import h5py
import megham.transform as mt
import megham.utils as mu
import numpy as np
import yaml
from scipy.cluster import vq
from scipy.optimize import minimize
from so3g.proj import quat
from sotodlib.coords import optics as op
from sotodlib.coords.fp_containers import (
    FocalPlane,
    OpticsTube,
    Receiver,
    Template,
    Transform,
    plot_by_gamma,
    plot_ot,
    plot_receiver,
    plot_ufm,
)
from sotodlib.coords.pointing_model import apply_pointing_model
from sotodlib.core import AxisManager, Context, IndexAxis, metadata
from sotodlib.io.metadata import read_dataset
from sotodlib.site_pipeline import util

logger = util.init_logger(__name__, "finalize_focal_plane: ")


def _create_db(filename, per_obs, obs_ids, start_time, stop_time):
    if per_obs:
        if len(obs_ids) != 1:
            raise ValueError(f"Running in per_obs mode but {len(obs_ids)} found!")
        base = {"obs:obs_id": obs_ids[0]}
        group = obs_ids[0]
    else:
        base = {"obs:timestamp": (start_time, stop_time)}
        group = str(start_time)
    if os.path.isfile(filename):
        return metadata.ManifestDb(filename), base, group
    if not os.path.isdir(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename), exist_ok=True)

    scheme = metadata.ManifestScheme()
    scheme.add_exact_match("dets:stream_id")
    if per_obs:
        scheme.add_exact_match("obs:obs_id")
    else:
        scheme.add_range_match("obs:timestamp")
    scheme.add_data_field("dataset")

    metadata.ManifestDb(scheme=scheme).to_file(filename)
    return metadata.ManifestDb(filename), base, group


def _avg_focalplane(full_fp, tot_weight):
    # Figure out how many good pointings we have for each det
    msk = np.isfinite(full_fp)
    n_obs = np.sum(np.any(msk, axis=1), axis=-1)
    n_point, _, n_gamma = tuple(np.sum(msk, axis=-1).T)
    tot_weight[tot_weight[:, 0] == 0] = np.nan
    avg_fp = np.nansum(full_fp, axis=-1) / tot_weight[:, 0][..., None]
    avg_weight = tot_weight / n_obs[..., None]

    # nansum all all nans is 0, addressing that case here
    all_nan = ~np.any(
        np.isfinite(np.swapaxes(full_fp, 0, 1)).reshape((full_fp.shape[1], -1)), axis=1
    )
    avg_fp[:, all_nan] = np.nan

    return avg_fp, avg_weight, n_point, n_gamma


def _log_vals(shift, scale, shear, rot, axis):
    deg2rad = np.pi / 180.0
    rad2deg = 180.0 / np.pi
    for ax, s in zip(axis, shift):
        logger.info("\tShift along %s axis is %f", ax, s)
    for ax, s in zip(axis, scale):
        logger.info("\tScale along %s axis is %f", ax, s)
        if np.isclose(s, deg2rad):
            logger.warning(
                "\tScale factor for %s looks like a degrees to radians conversion", ax
            )
        elif np.isclose(s, rad2deg):
            logger.warning(
                "\tScale factor for %s looks like a radians to degrees conversion", ax
            )
    logger.info("\tShear param is %f", shear)
    logger.info("\tRotation of the %s-%s plane is %f radians", axis[0], axis[1], rot)


def gamma_fit(src, dst):
    """
    Fit the transformation for gamma.
    Note that the periodicity here assumes things are in radians.

    Arguments:

        src: Source gamma in radians

        dst: Destination gamma in radians

    Returns:

       scale: Scale applied to src

       shift: Shift applied to scale*src
    """

    def _gamma_min(pars, src, dst):
        scale, shift = pars
        transformed = np.sin(src * scale + shift)
        diff = np.sin(dst) - transformed

        return np.sqrt(np.mean(diff**2))

    res = minimize(_gamma_min, (1.0, 0.0), (src, dst))
    return res.x


def _load_template(template_path, ufm, pointing_cfg):
    template_rset = read_dataset(template_path, ufm)
    det_ids = template_rset["dets:det_id"]
    template = np.column_stack(
        (
            np.array(template_rset["xi"]),
            np.array(template_rset["eta"]),
            np.array(template_rset["gamma"]),
        )
    )
    template_optical = template_rset["is_optical"]

    return Template(
        np.array(det_ids), template, np.array(template_optical), pointing_cfg
    )


def _get_obs_ids(
    ctx,
    metalist,
    start_time,
    stop_time,
    query=None,
    obs_ids=[],
    tags=[],
    stream_ids=[],
    min_dets=0,
):
    all_obs = obs_ids
    query_obs = []
    if len(obs_ids) == 0:
        query_all = query
        if query is None:
            query_all = (
                f"type=='obs' and start_time>{start_time} and stop_time<{stop_time}"
            )
        if ctx.obsdb is None:
            raise ValueError("No obsdb!")
        all_obs = ctx.obsdb.query(query_all, tags=tags)["obs_id"]
    dbs = [
        metadata.ManifestDb(md["db"])
        for md in ctx["metadata"]
        if md.get("name", "") in metalist or md.get("label", "") in metalist
    ]
    with_meta = np.unique(
        np.hstack(
            [np.array([entry["obs:obs_id"] for entry in db.inspect()]) for db in dbs]
        )
    )
    all_obs = np.intersect1d(all_obs, with_meta)

    if query is not None:
        query_obs = ctx.obsdb.query(query)["obs_id"]
    obs_ids += query_obs

    if len(stream_ids) > 0:
        all_obs = [
            obs_id
            for obs_id in all_obs
            if len(
                np.intersect1d(
                    ctx.obsdb.get(obs_id)["stream_ids_list"].split(","), stream_ids
                )
            )
        ]
    all_obs = np.array(
        [
            obs_id
            for obs_id in all_obs
            if len(ctx.obsfiledb.get_det_table(obs_id)) >= min_dets
        ]
    )

    if len(obs_ids) == 0 and query is None:
        return all_obs
    return np.intersect1d(obs_ids, all_obs)


def _load_ctx(config):
    ctx = Context(config["context"]["path"])
    if ctx.obsdb is None:
        raise ValueError("No obsdb!")
    tod_pointing_name = config["context"].get("tod_pointing", "tod_pointing")
    map_pointing_name = config["context"].get("map_pointing", "map_pointing")
    pol_name = config["context"].get("polarization", "polarization")
    dm_name = config["context"].get("detmap", "detmap")
    roll_range = config.get("roll_range", [-1 * np.inf, np.inf])
    obs_ids = _get_obs_ids(
        ctx,
        [tod_pointing_name, map_pointing_name, pol_name],
        config["start_time"],
        config["stop_time"],
        config["context"].get("query", None),
        config["context"].get("obs_ids", []),
        config["context"].get("tags", ["timing_issues=0"]),
        config["context"].get("stream_ids", []),
        config.get("min_points", 0),
    )
    if len(obs_ids) == 0:
        logger.warning("No observations provided in configuration")
    amans = []
    dets = config["context"].get("dets", {})
    for obs_id in obs_ids:
        roll = ctx.obsdb.get(obs_id)["roll_center"]
        if roll is None:
            continue
        if roll < roll_range[0] or roll > roll_range[1]:
            logger.info("%s has a roll that is out of range", obs_id)
            continue
        try:
            aman = ctx.get_meta(obs_id, dets=dets)
        except metadata.loader.LoaderError:
            logger.error("Failed to load %s, skipping", obs_id)
            continue
        if aman.obs_info.tube_slot == "stp1":
            aman.obs_info.tube_slot = "st1"
        if "det_info" not in aman:
            raise ValueError(f"No det_info in {obs_id}")
        if "wafer" not in aman.det_info and dm_name in aman:
            dm_aman = aman[dm_name].copy()
            aman.det_info.wrap("wafer", dm_aman)
            if "det_id" not in aman.det_info:
                aman.det_info.wrap(
                    "det_id", aman.det_info.wafer.det_id, [(0, aman.dets)]
                )
        if "det_id" in aman.det_info:
            aman.restrict("dets", ~np.isin(aman.det_info.det_id, ["", "NO_MATCH"]))
        else:
            raise ValueError(f"No detmap for {obs_id}")
        pol = pol_name in aman
        if pol:
            aman.move(pol_name, "polarization")
        else:
            logger.warning("No polarization data in context")

        if tod_pointing_name in aman:
            _aman = aman.copy()
            _aman.move(tod_pointing_name, "pointing")
            amans.append(_aman)
        if map_pointing_name in aman:
            _aman = aman.copy()
            _aman.move(map_pointing_name, "pointing")
            amans.append(_aman)
        elif tod_pointing_name not in aman:
            raise ValueError(f"No pointing found in {obs_id}")
    obs_ids = [aman.obs_info.obs_id for aman in amans]

    # Figure out stream_ids and OTs
    query_all = f"type=='obs' and start_time>{config['start_time']} and stop_time<{config['stop_time']}"
    ot_sid = []
    for obs in ctx.obsdb.query(query_all):
        if obs["wafer_slots_list"] is None or obs["stream_ids_list"] is None:
            continue
        ot = obs["tube_slot"]
        if ot == "stp1":
            ot = "st1"
        ot_sid += [
            (obs["telescope_flavor"], ot, ws, sid)
            for ws, sid in zip(
                obs["wafer_slots_list"].split(","), obs["stream_ids_list"].split(",")
            )
        ]
    ot_sid = np.unique(np.array(ot_sid), axis=0)

    return amans, obs_ids, ot_sid


def _load_rset_single(config):
    obs_id = config["resultsets"].get("obs_id", "")
    pointing_rset = read_dataset(*config["resultsets"]["pointing"])
    pointing_aman = pointing_rset.to_axismanager(axis_key="dets:readout_id")
    aman = AxisManager(pointing_aman.dets)
    aman = aman.wrap("pointing", pointing_aman)

    if "polarization" in config["resultsets"]:
        polarization_rset = read_dataset(*config["resultsets"]["polarization"])
        polarization_aman = polarization_rset.to_axismanager(axis_key="dets:readout_id")
        aman = aman.wrap("polarization", polarization_aman)

    det_info = AxisManager(aman.dets)
    dm_rset = read_dataset(*config["resultsets"]["detmap"])
    dm_aman = dm_rset.to_axismanager(axis_key="readout_id")
    det_info.wrap("wafer", dm_aman)
    det_info.wrap("readout_id", det_info.dets.vals, [(0, det_info.dets)])
    det_info.wrap("det_id", det_info.wafer.det_id, [(0, det_info.dets)])
    det_info.wrap(
        "stream_id",
        np.array([config["stream_id"].lower()] * det_info.dets.count),
        [(0, det_info.dets)],
    )
    det_info.wrap(
        "wafer_slot",
        np.array([config["wafer_slot"].lower()] * det_info.dets.count),
        [(0, det_info.dets)],
    )
    det_info.restrict("dets", det_info.dets.vals[det_info.det_id != ""])
    det_info.det_id = np.char.strip(det_info.det_id)  # Needed for some old results
    aman = aman.wrap("det_info", det_info)
    aman.restrict("dets", aman.dets.vals[aman.det_info.det_id != "NO_MATCH"])

    obs_info = AxisManager()
    obs_info.wrap("telescope_flavor", config["telescope_flavor"].lower())
    obs_info.wrap("tube_slot", config["tube_slot"].lower())
    aman.wrap("obs_info", obs_info)

    smurf = AxisManager(aman.dets)
    if "band" in aman.pointing:
        smurf.wrap("band", np.array(aman.pointing.band, dtype=int), [(0, smurf.dets)])
    elif "wafer" in det_info and "smurf_band" in det_info.wafer:
        smurf.wrap(
            "band", np.array(det_info.wafer.smurf_band, dtype=int), [(0, smurf.dets)]
        )
    if "channel" in aman.pointing:
        smurf.wrap(
            "channel", np.array(aman.pointing.channel, dtype=int), [(0, smurf.dets)]
        )
    elif "wafer" in det_info and "smurf_channel" in det_info.wafer:
        smurf.wrap(
            "channel",
            np.array(det_info.wafer.smurf_channel, dtype=int),
            [(0, smurf.dets)],
        )
    aman.det_info.wrap("smurf", smurf)

    return aman, obs_id


def _load_rset(config):
    stream_id = config["stream_id"]
    telescope_flavor = config["telescope_flavor"].lower()
    ot = config["tube_slot"].lower()
    ws = config["wafer_slot"].lower()
    obs = config["resultsets"]
    _config = config.copy()
    obs_ids = np.array(list(obs.keys()))
    amans: List[Optional[AxisManager]] = [None] * len(obs_ids)
    obs_info = AxisManager()
    obs_info.wrap("stream_id", stream_id)
    for i, (obs_id, rsets) in enumerate(obs.items()):
        _config["resultsets"] = rsets
        _config["resultsets"]["obs_id"] = obs_id
        aman, _ = _load_rset_single(_config)
        if "det_info" not in aman or "det_id" not in aman.det_info:
            raise ValueError(f"No detmap for {obs_id}")
        amans[i] = aman

    return (
        amans,
        obs_ids,
        [(telescope_flavor, ot, ws, stream_id)],
    )


def _mk_pointing_config(telescope_flavor, tube_slot, wafer_slot, config):
    config_dir = config.get("pipeline_config_dir", os.environ["PIPELINE_CONFIG_DIR"])
    config_path = os.path.join(config_dir, "shared/focalplane/ufm_to_fp.yaml")
    zemax_path = config.get("zemax_path", None)

    pointing_cfg = {
        "telescope_flavor": telescope_flavor,
        "tube_slot": tube_slot,
        "wafer_slot": wafer_slot,
        "config_path": config_path,
        "zemax_path": zemax_path,
        "return_fp": False,
    }
    return pointing_cfg


def _restrict_inliers(aman, focal_plane):
    # TODO: Use gamma as well
    # Map to template
    fp, _, template_msk = focal_plane.map_by_det_id(aman)
    fp = fp[:, :2]
    inliers = np.ones(len(fp), dtype=bool)

    rad_thresh = 1.05 * np.nanmax(
        np.linalg.norm(
            focal_plane.template.fp[:, :2] - focal_plane.template.center[:, :2], axis=1
        )
    )

    # Use kmeans to kill any ghosts
    fp_white = vq.whiten(fp[inliers])
    codebook, _ = vq.kmeans(fp_white, 2)
    codes, _ = vq.vq(fp_white, codebook)

    c0 = codes == 0
    c1 = codes == 1
    m0 = np.median(fp[inliers][c0], axis=0)
    m1 = np.median(fp[inliers][c1], axis=0)
    dist = np.linalg.norm(m0 - m1)

    # If centroids are too far from each other use the bigger one
    if dist < rad_thresh:
        cluster = c0 + c1
    elif np.sum(c0) >= np.sum(c1):
        cluster = c0
    else:
        cluster = c1

    # Flag anything too far away from the center
    cent = np.median(fp[inliers][cluster], axis=0)
    r = np.linalg.norm(fp[inliers] - cent, axis=1)
    inliers[inliers] *= cluster * (r <= rad_thresh)

    # Now kill dets that seem too far from their match
    fp[~inliers] = np.nan
    rot, sft = mt.get_rigid(fp, focal_plane.template.fp[template_msk, :2])
    fp_aligned = mt.apply_transform(fp, rot, sft)
    likelihood = mu.gen_weights(fp_aligned, focal_plane.template.fp[template_msk, :2])
    inliers *= likelihood > 0.61  # ~1 sigma cut

    # Now restrict the AxisManager
    inlier_det_ids = focal_plane.template.det_ids[template_msk][inliers]
    return aman.restrict(
        "dets", aman.dets.vals[np.isin(aman.det_info.det_id, inlier_det_ids)]
    )


def _apply_pointing_model(config, aman):
    if "pointing_model" not in config:
        logger.info("\t\tNo pointing model specified!")
        return aman
    if not config["pointing_model"].get("apply", False):
        logger.info("\t\tNot applying pointing model")
        return aman
    if "function" not in config["pointing_model"]:
        logger.info("\t\tUsing default pointing model function")
        func = apply_pointing_model
    else:
        func = getattr(
            import_module(config["pointing_model"]["function"][0]),
            config["pointing_model"]["function"][1],
        )
    if "az" not in aman.pointing:
        raise ValueError("Need to have az in pointing fits to apply pointing model")
    if "el" not in aman.pointing:
        raise ValueError("Need to have el in pointing fits to apply pointing model")
    if "roll" not in aman.pointing:
        raise ValueError("Need to have roll in pointing fits to apply pointing model")

    params = config["pointing_model"].get("params", {})
    if "pointing_model" in aman:
        for key, val in params.items():
            if key in aman.pointing_model:
                aman.pointing_model[key] = val
            else:
                aman.pointing_model.wrap(key, val)
        params = aman.pointing_model
    ancil = AxisManager(IndexAxis("samps", aman.dets.count))
    ancil.wrap("az_enc", np.rad2deg(aman.pointing.az))
    ancil.wrap("el_enc", np.rad2deg(aman.pointing.el))
    ancil.wrap("roll_enc", np.rad2deg(aman.pointing.roll))
    bs = func(aman, params, ancil, False)
    q_fp = quat.rotation_xieta(aman.pointing.xi, aman.pointing.eta)
    have_gamma = False
    if "gamma" in aman.pointing:
        if np.any(np.isnan(aman.pointing.gamma)):
            logger.warning(
                "\t\tnans in gamma, not including in pointing model correction"
            )
        else:
            q_fp = quat.rotation_xieta(
                aman.pointing.xi, aman.pointing.eta, aman.pointing.gamma
            )
            have_gamma = True

    xi, eta, gamma = quat.decompose_xieta(
        ~quat.euler(2, bs.roll)
        * ~quat.rotation_lonlat(-bs.az, bs.el)
        * quat.rotation_lonlat(-1 * aman.pointing.az, aman.pointing.el)
        * quat.euler(2, aman.pointing.roll)
        * q_fp
    )

    aman.pointing.xi[:] = xi
    aman.pointing.eta[:] = eta
    if have_gamma:
        aman.pointing.gamma[:] = gamma

    return aman


def _reverse_roll(fp, aff, sft, aman):
    if "obs_info" not in aman:
        raise ValueError("Can't reverse roll without obs information")
    if "roll_center" not in aman.obs_info:
        raise ValueError("Can't reverse roll without roll information")
    roll = -1 * np.deg2rad(aman.obs_info.roll_center)

    # We want to shift so we rotating about the origin
    # To get to nominal we do fp@aff + sft
    # So if we just want to recenter we do fp + sft@aff^-1
    inv_aff, _ = mt.invert_transform(aff, np.zeros_like(sft))
    sft_adj = sft @ inv_aff
    fp_sft = fp[:, :2] + sft_adj

    # Now lets reverse the roll
    # The transpose is the inverse
    rot = np.array([[np.cos(roll), -1 * np.sin(roll)], [np.sin(roll), np.cos(roll)]])
    fp_rot = fp_sft @ rot

    # And undo the shift, keeping track of rotations
    fp_rot -= sft_adj @ rot

    # Make sure its set
    fp[:, :2] = fp_rot

    # For gamma lets just shift by the roll
    fp[:, 2] -= roll

    return fp


def main():
    # Read in input pars
    parser = ap.ArgumentParser()
    parser.add_argument("config_path", help="Location of the config file")
    parser.add_argument(
        "--per_obs", "-p", action="store_true", help="Run in per observation mode"
    )
    parser.add_argument(
        "--include_cm",
        "-i",
        action="store_true",
        help="Include the common mode in the final detector positions",
    )
    args = parser.parse_args()

    # Open config file
    with open(args.config_path, "r", encoding="utf-8") as file:
        config = yaml.safe_load(file)
    per_obs = config.get("per_obs", args.per_obs)
    include_cm = config.get("include_cm", args.include_cm)

    # Build output path
    append = config.get("append", "")
    dbroot = f"db{bool(append)*'_'}{append}"
    froot = f"focal_plane{bool(append)*'_'}{append}"
    subdir = config.get("subdir", "")
    subdir = subdir + (subdir == "") * (
        per_obs * "per_obs" + (not per_obs) * "combined"
    )
    outdir = os.path.join(config["outdir"], subdir)
    outpath = os.path.abspath(os.path.join(outdir, f"{froot}.h5"))
    dbpath = os.path.join(outdir, f"{dbroot}.sqlite")
    logpath = os.path.join(outdir, f"{froot}.log")
    os.makedirs(outdir, exist_ok=True)
    plot_dir_base = config.get("plot_dir", None)
    if plot_dir_base is not None:
        plot_dir_base = os.path.join(
            plot_dir_base, subdir + bool(append) * "_" + append
        )
        plot_dir_base = os.path.abspath(plot_dir_base)
        os.makedirs(plot_dir_base, exist_ok=True)

    # Log file
    logfile = logging.FileHandler(logpath)
    logger.addHandler(logfile)

    # Time range
    config["start_time"] = config.get("start_time", 0)
    config["stop_time"] = config.get("stop_time", 2**32)
    logger.info(
        "Running on time range %s to %s",
        dt.datetime.fromtimestamp(config["start_time"]),
        dt.datetime.fromtimestamp(config["stop_time"]),
    )

    # Load data
    if "context" in config:
        amans, obs_ids, ot_sids = _load_ctx(config)
    elif "resultsets" in config:
        amans, obs_ids, ot_sids = _load_rset(config)
    else:
        raise ValueError("No valid inputs provided")
    if len(ot_sids) == 0:
        raise ValueError("No stream_ids found!")
    if np.any(ot_sids[:, 0] != ot_sids[0][0]):
        raise ValueError("Not all AxisManagers agree on telescope!")

    weight_factor = config.get("weight_factor", 1000)
    min_points = config.get("min_points", 50)
    gen_template = "template" not in config
    template_path = config.get("template", "nominal.h5")
    have_template = os.path.exists(template_path)
    if not gen_template and not have_template:
        logger.error("Provided template doesn't exist, trying to generate one")
        gen_template = True

    # Serialize config
    repo = git.Repo(
        os.path.abspath(os.path.dirname(__file__)), search_parent_directories=True
    )
    sha = f"{repo.head.object.hexsha}{'_dirty'*repo.is_dirty()}"
    config["git_sha"] = sha
    cfg_str = str(yaml.dump(config))

    # Need to move installed OT and WS of array to templace for this
    # if config.get("pad", False):
    #     logger.info("Padding missing arrays with template, getting complete list of arrays from template")
    #     if not have_template:
    #         logger.warning("\tNo template provided, arrays not found in any observations will be missing")
    #     with h5py.File(template_path) as f:
    #         stream_ids = list(f.keys())

    # Split up into batches
    # Right now either per_obs or all at once
    # Maybe allow for batch my encoder angle later?
    if per_obs:
        logger.info("Running in per_obs mode")
        batches = [([aman], [obs_id]) for aman, obs_id in zip(amans, obs_ids)]
    else:
        batches = [(amans, obs_ids)]
    stream_ids = config.get("context", {}).get("stream_ids", [])
    for amans, obs_ids in batches:
        plot_dir = plot_dir_base
        if per_obs:
            plot_dir = os.path.join(plot_dir_base, obs_ids[0])
        else:
            plot_dir = os.path.join(plot_dir_base, str(config["start_time"]))
        os.makedirs(plot_dir, exist_ok=True)
        logger.info("Working on batch containing: %s", str(obs_ids))

        # Setup db and Receiver
        db, base, group = _create_db(
            dbpath,
            per_obs=per_obs,
            obs_ids=obs_ids,
            start_time=config["start_time"],
            stop_time=config["stop_time"],
        )
        rx = Receiver()
        if config.get("in_place", True):
            with h5py.File(outpath, "a") as f:
                if group in f:
                    rx = Receiver.load(f, group)
        for tel, ot, ws, stream_id in ot_sids:
            if len(stream_ids) > 0 and stream_id not in stream_ids:
                continue
            logger.info("Working on %s", stream_id)

            # Limit ourselves to amans with this stream_id and restrict
            amans_restrict = [
                aman.copy().restrict(
                    "dets", aman.dets.vals[aman.det_info.stream_id == stream_id]
                )
                for aman in amans
                if aman is not None and stream_id in aman.det_info.stream_id
            ]
            obs_ids_restrict = [
                obs_id
                for aman, obs_id in zip(amans, obs_ids)
                if aman is not None and stream_id in aman.det_info.stream_id
            ]
            if len(amans_restrict) == 0:
                message = "\tSomehow no AxisManagers with stream_id %s"
                if per_obs:
                    logger.info(message, stream_id)
                    continue
                else:
                    logger.error(message, stream_id)

            # Make pointing config
            logger.info("\t%s is in %s %s %s", stream_id, tel, ot, ws)
            pointing_cfg = _mk_pointing_config(tel, ot, ws, config)

            # Cnstructing the OT if we need to
            if ot in rx.ot_dict:
                rx.ot_dict[ot].delete_fp(stream_id)
            else:
                optics_tube = OpticsTube.from_pointing_cfg(pointing_cfg)
                rx.optics_tubes = rx.optics_tubes + [optics_tube]

            # If a template is provided load it, otherwise generate one
            if gen_template:
                logger.info(f"\tGenerating template for {stream_id}")
                if "wafer_info" not in config:
                    raise ValueError("Need wafer_info to generate template")
                template_det_ids, template, is_optical = op.gen_template(
                    config["wafer_info"], stream_id, **pointing_cfg
                )
                template = Template(
                    template_det_ids,
                    template,
                    is_optical,
                    pointing_cfg,
                )
            elif have_template:
                logger.info("\tLoading template from %s", template_path)
                template = _load_template(template_path, stream_id, pointing_cfg)
            else:
                raise ValueError(
                    "No template provided and unable to generate one for some reason"
                )

            focal_plane = FocalPlane.empty(
                template, stream_id, ws, len(amans), config=cfg_str
            )
            if focal_plane.template is None:
                raise ValueError("Template is somehow None")

            n_obs = 0
            for i, (aman, obs_id) in enumerate(zip(amans_restrict, obs_ids_restrict)):
                logger.info("\tWorking on %s", obs_id)
                if aman.dets.count < min_points:
                    logger.info("\t\tToo few dets found, skipping")
                    continue

                if config.get("faked_gamma", False):
                    aman.pointing.gamma[:] = np.nan

                # Restrict to optical dets
                optical = np.isin(
                    aman.det_info.det_id, focal_plane.template.det_ids[template.optical]
                )
                aman.restrict("dets", aman.dets.vals[optical])
                if aman.dets.count == 0:
                    logger.info("\t\tNo optical dets, skipping...")
                    continue

                # Apply pointing model if we want to
                aman = _apply_pointing_model(config, aman)

                # Do some outlier cuts
                if "hits" in aman.pointing:
                    aman.restrict(
                        "dets", aman.pointing.hits >= config.get("min_hits", 5)
                    )
                    if aman.dets.count == 0:
                        logger.info("\t\tNo high hits dets, skipping...")
                        continue
                if "R2" in aman.pointing:
                    aman.restrict("dets", aman.pointing.R2 > config.get("min_r2", 0.7))
                    if aman.dets.count == 0:
                        logger.info("\t\tNo high R2 dets, skipping...")
                        continue
                if aman.dets.count < min_points:
                    logger.info("\t\tToo few dets found, skipping")
                    continue
                _restrict_inliers(aman, focal_plane)

                # Mapping to template
                fp, r2, template_msk = focal_plane.map_by_det_id(aman)
                focal_plane.template.add_wafer_info(aman, template_msk)

                # Try an initial alignment and get weights
                try:
                    aff, sft = mt.get_rigid(
                        fp[:, :2], focal_plane.template.fp[template_msk, :2]
                    )
                except ValueError as e:
                    logger.error("\t\t%s", e)
                    continue
                aligned = mt.apply_transform(fp[:, :2], aff, sft)

                if config.get("reverse_roll", False):
                    fp = _reverse_roll(fp, aff, sft, aman)
                if np.any(np.isfinite(fp[:, 2])):
                    gscale, gsft = gamma_fit(
                        fp[:, 2], focal_plane.template.fp[template_msk, 2]
                    )
                    weights = mu.gen_weights(
                        np.column_stack((aligned, gscale * fp[:, 2] + gsft)),
                        focal_plane.template.fp[template_msk],
                        focal_plane.template.spacing.ravel() / weight_factor,
                    )
                else:
                    weights = mu.gen_weights(
                        aligned,
                        focal_plane.template.fp[template_msk, :2],
                        focal_plane.template.spacing[:2].ravel() / weight_factor,
                    )
                # ~1 sigma cut
                weights[weights < 0.61] = np.nan
                if np.sum(np.isfinite(weights)) < min_points / 2:
                    logger.error("\t\tToo few points! Skipping...")

                # Store weighted values
                weights = np.column_stack((weights, r2))
                focal_plane.add_fp(i, fp, weights, template_msk)

                n_obs += 1

            # Compute the average focal plane with weights
            (
                focal_plane.avg_fp,
                focal_plane.weights,
                focal_plane.n_point,
                focal_plane.n_gamma,
            ) = _avg_focalplane(focal_plane.full_fp, focal_plane.tot_weight)
            tot_points = np.sum((focal_plane.n_point > 0).astype(int))
            focal_plane.id_strs = focal_plane.template.id_strs
            logger.info("\t%d points from %d obs in fit", tot_points, n_obs)
            if tot_points < min_points:
                logger.error("\tToo few points! Skipping...")
                if config.get("pad", False):
                    logger.info("\tPadding output with template")
                    focal_plane.transformed = focal_plane.template.fp
                    focal_plane.tot_weight = None
                    rx.ot_dict[ot].focal_planes = rx.ot_dict[ot].focal_planes + [
                        focal_plane
                    ]
                continue

            try:
                affine, shift = mt.get_affine_two_stage(
                    focal_plane.template.fp[:, :2],
                    focal_plane.avg_fp[:, :2],
                    focal_plane.weights[:, 0],
                )
            except ValueError as e:
                logger.error("\t%s", e)
                continue

            focal_plane.transformed[:, :2] = mt.apply_transform(
                focal_plane.template.fp[:, :2], affine, shift
            )
            focal_plane.center_transformed[:, :2] = mt.apply_transform(
                focal_plane.template.center[:, :2], affine, shift
            )

            # Compute transformation between the two nominal and measured pointing
            focal_plane.have_gamma = np.sum(focal_plane.n_gamma) > 0
            if focal_plane.have_gamma:
                gamma_scale, gamma_shift = gamma_fit(
                    focal_plane.template.fp[:, 2], focal_plane.avg_fp[:, 2]
                )
            else:
                logger.warning(
                    "\tNo polarization data availible, gammas will be based on the nominal values."
                )
                logger.warning(
                    "\tSetting gamma shift to the xi-eta rotation and scale to 1.0"
                )
                transform = Transform.from_split(np.array((*shift, 0.0)), affine, 1.0)
                gamma_scale = 1.0
                gamma_shift = transform.rot
            focal_plane.transformed[:, 2] = (
                focal_plane.template.fp[:, 2] * gamma_scale + gamma_shift
            )
            focal_plane.center_transformed[:, 2] = (
                focal_plane.template.center[:, 2] * gamma_scale + gamma_shift
            )

            rms = np.sqrt(
                np.nanmean(
                    (
                        focal_plane.avg_fp[:, : (2 + focal_plane.have_gamma)]
                        - focal_plane.transformed[:, : (2 + focal_plane.have_gamma)]
                    )
                    ** 2
                )
            )
            logger.info("\tRMS after transformation is %f", rms)

            shift = np.array((*shift, gamma_shift))
            focal_plane.transform = Transform.from_split(shift, affine, gamma_scale)
            _log_vals(
                focal_plane.transform.shift,
                focal_plane.transform.scale,
                focal_plane.transform.shear,
                focal_plane.transform.rot,
                ("xi", "eta", "gamma"),
            )

            if config.get("plot", False):
                plot_ufm(focal_plane, plot_dir)
                plot_by_gamma(focal_plane, plot_dir)

            # Add to the receiver
            rx.ot_dict[ot].focal_planes = rx.ot_dict[ot].focal_planes + [focal_plane]

        # Per OT common mode
        todel = []
        for name, ot in rx.ot_dict.items():
            logger.info("Fitting common mode for %s", ot.name)
            centers = np.atleast_2d(
                np.array(
                    [
                        fp.template_center
                        for fp in ot.focal_planes
                        if fp.tot_weight is not None
                    ]
                )
            )
            centers_transformed = np.atleast_2d(
                np.array(
                    [
                        fp.center_transformed
                        for fp in ot.focal_planes
                        if fp.tot_weight is not None
                    ]
                )
            )
            if ot.num_fps == 0 or centers.size == 0 or centers_transformed.size == 0:
                logger.error("\tNo focal planes found! Skipping...")
                if not config.get("pad", False):
                    todel.append(name)
                continue
            plot_ot(ot, plot_dir)
            centers = centers.reshape((-1, 3))
            centers_transformed = centers_transformed.reshape((-1, 3))
            if centers.shape[0] < 3:
                logger.warning(
                    "\tToo few wafers fit to compute common mode, transform will be approximated"
                )
                centers = np.vstack([ot.center, ot.center - 1, ot.center + 1])
                centers_transformed = np.vstack(
                    [
                        mt.apply_transform(
                            centers, fp.transform.affine, fp.transform.shift
                        )
                        for fp in ot.focal_planes
                    ],
                )
                centers = np.repeat(centers, len(ot.focal_planes), 0)
            rot, sft = mt.get_rigid(centers[:, :2], centers_transformed[:, :2])
            gamma_shift = np.mean(centers_transformed[:, 2] - centers[:, 2])
            ot.transform_fullcm = Transform.from_split(
                np.array((*sft.ravel(), gamma_shift)), rot, 1.0
            )
            ot.center_transformed = mt.apply_transform(
                ot.center, ot.transform_fullcm.affine, ot.transform_fullcm.shift
            )
            _log_vals(
                ot.transform_fullcm.shift,
                ot.transform_fullcm.scale,
                ot.transform_fullcm.shear,
                ot.transform_fullcm.rot,
                ("xi", "eta", "gamma"),
            )
        logger.info("Deleting OTs: %s", str(todel))
        for ot in todel:
            rx.delete_ot(ot)

        # Full receiver common mode
        logger.info("Fitting receiver common mode")
        centers = np.atleast_2d(
            np.array([ot.center for ot in rx.optics_tubes if ot.num_fps > 0])
        )
        centers_transformed = np.atleast_2d(
            np.array(
                [ot.center_transformed for ot in rx.optics_tubes if ot.num_fps > 0]
            )
        )
        no_ots = (
            len(rx.optics_tubes) == 0
            or centers.size == 0
            or centers_transformed.size == 0
        )
        if no_ots and not config.get("pad", False):
            logger.error("\tNo optics tubes found! Skipping...")
            continue
        elif len(rx.optics_tubes) == 1 or (no_ots and config.get("pad", False)):
            logger.info(
                "\tOnly one OT found, receiver common mode will be from this tube"
            )
            recv_transform = deepcopy(rx.optics_tubes[0].transform_fullcm)
        else:
            centers = centers.reshape((-1, 3))
            centers_transformed = centers_transformed.reshape((-1, 3))
            if len(rx.optics_tubes) < 3:
                logger.info(
                    "\tNot enough OTs to fit receiver common mode, transform will be approximated"
                )
                centers = np.vstack(
                    [np.roll(np.arange(3, dtype=float), i) for i in range(3)]
                )
                centers_transformed = np.vstack(
                    [
                        mt.apply_transform(
                            centers, ot.transform.affine, ot.transform.shift
                        )
                        for ot in rx.optics_tubes
                    ],
                )
                centers = np.repeat(centers, len(rx.optics_tubes), 0)
            rot, sft = mt.get_rigid(centers[:, :2], centers_transformed[:, :2])
            gamma_shift = np.mean(centers_transformed[:, 2] - centers[:, 2])
            recv_transform = Transform.from_split(
                np.array((*sft.ravel(), gamma_shift)), rot, 1.0
            )
        receiver = Receiver(
            rx.optics_tubes,
            transform=recv_transform,
            include_cm=include_cm,
            valid_ids=np.unique(rx.valid_ids + rx.fp_valid_ids).tolist(),
        )

        if not (no_ots and config.get("pad", False)):
            _log_vals(
                recv_transform.shift,
                recv_transform.scale,
                recv_transform.shear,
                recv_transform.rot,
                ("xi", "eta", "gamma"),
            )
            plot_receiver(receiver, plot_dir)

        # Now compute correction only transform for each ufm
        # Transforms are composed as ufm(ot(rx(focal_plane)))
        for ot in receiver.optics_tubes:
            # Now remove the receiver CM from the OT
            ot.transform.affine, ot.transform.shift = mt.decompose_transform(
                ot.transform_fullcm.affine,
                ot.transform_fullcm.shift,
                recv_transform.affine,
                recv_transform.shift,
            )
            ot.transform.decompose()

            # Now for each fp remove the CM
            for fp in ot.focal_planes:
                (
                    fp.transform_nocm.affine,
                    fp.transform_nocm.shift,
                ) = mt.decompose_transform(
                    fp.transform.affine,
                    fp.transform.shift,
                    ot.transform_fullcm.affine,
                    ot.transform_fullcm.shift,
                )
                fp.transform_nocm.decompose()
                # Remove the common mode if desired
                if not include_cm and fp.template is not None:
                    fp.with_cm = False
                    fp.transformed = mt.apply_transform(
                        fp.template_fp,
                        fp.transform_nocm.affine,
                        fp.transform_nocm.shift,
                    )

        # Make final outputs and save
        logger.info("Saving data to %s", outpath)
        logger.info("Writing to database at %s", dbpath)
        if config.get("pad", False):
            logger.info("Padding missing arrays with values from template")
        with h5py.File(outpath, "a") as f:
            if group in f:
                del f[group]
            f.create_group(group)
            receiver.save(f, (db, base), group)


if __name__ == "__main__":
    main()



# Copyright (c) 2020-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import re

import numpy as np
import traitlets
from astropy import units as u
import healpy as hp

from toast.timing import function_timer, Timer
from toast import qarray as qa
from toast.traits import trait_docs, Int, Unicode, Bool
from toast.ops.operator import Operator
from toast.utils import Logger
from toast.observation import default_values as defaults

from ...core.hardware import LAT_COROTATOR_OFFSET


XAXIS, YAXIS, ZAXIS = np.eye(3)

@trait_docs
class CoRotator(Operator):
    """Compute the LAT focalplane rotation.

    The optical design configuration of the LAT projected on the sky has the "O6" cryo
    tube in the positive azimuth direction when the telescope is at 60 degrees
    elevation.

    If corotate_lat is False, then the projected focalplane will rotate on the sky.
    The saved corotation angle will be zero.  If corotate_lat is True, then the
    corotation angle is set to compensate for the rotation caused by the observing
    elevation.

    The rotation of the projected focalplane is given by:

        R = elevation - offset - corotator angle

    This operator should be applied *before* expanding the detector pointing
    from boresight.

    """

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    boresight_azel = Unicode(
        defaults.boresight_azel,
        allow_none=True,
        help="Observation shared key for boresight Az/El",
    )

    boresight_radec = Unicode(
        defaults.boresight_radec,
        allow_none=True,
        help="Observation shared key for boresight RA/Dec",
    )

    corotator_angle = Unicode(
        "corotator_angle",
        allow_none=True,
        help="Observation shared key for corotation angle",
    )

    elevation = Unicode(
        defaults.elevation,
        allow_none=True,
        help="Observation shared key for boresight elevation",
    )

    corotate_lat = Bool(
        True,
        help="If True, rotate LAT receiver to maintain projected focalplane orientation"
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()
        timer = Timer()
        timer.start()

        for obs in data.obs:
            if re.match(r"LAT.*", obs.telescope.name) is None:
                msg = f"Skipping corotation for telescope '{obs.telescope.name}'"
                log.debug(msg)
                continue

            el_rad = obs.shared[self.elevation].data

            if self.corotator_angle not in obs.shared:
                obs.shared.create_column(
                    self.corotator_angle,
                    shape=(obs.n_local_samples,),
                    dtype=np.float64,
                )

            corot = None
            if self.corotate_lat:
                # We are corotating the receiver.  We set this to a fixed angle
                # based on the nominal elevation for the scan.
                if obs.comm_col_rank == 0:
                    if "scan_el" in obs:
                        scan_el = obs["scan_el"].to_value(u.radian)
                    else:
                        scan_el = np.mean(el_rad)
                    scan_el_deg = scan_el * 180.0 / np.pi
                    corot = np.zeros(obs.n_local_samples) + (
                        scan_el - LAT_COROTATOR_OFFSET.to_value(u.radian)
                    )
                    corot_deg = corot * 180.0 / np.pi
                    msg = f"LAT Co-rotation:  obs {obs.name} at scan El = "
                    msg += f"{scan_el_deg:0.2f} degrees, rotating by "
                    msg += f"{np.mean(corot_deg):0.2f} average degrees"
                    log.debug(msg)
                obs.shared[self.corotator_angle].set(
                    corot,
                    offset=(0,),
                    fromrank=0,
                )
            else:
                # We are not co-rotating.  Set the angle to zero.
                if obs.comm_col_rank == 0:
                    msg = f"LAT Co-rotation:  obs {obs.name} disabled"
                    log.debug(msg)
                    corot = np.zeros(obs.n_local_samples)
                obs.shared[self.corotator_angle].set(
                    corot,
                    offset=(0,),
                    fromrank=0,
                )

            # Now correct the boresight quaternions

            rot = None
            if obs.comm_col_rank == 0:
                rot = qa.from_axisangle(
                    ZAXIS,
                    el_rad - LAT_COROTATOR_OFFSET.to_value(u.rad) - obs.shared[self.corotator_angle].data,
                )

            for name in [self.boresight_radec, self.boresight_azel]:
                if name is None:
                    continue
                quats = obs.shared[name]
                new_quats = None
                if obs.comm_col_rank == 0:
                    new_quats = qa.mult(quats, rot)
                quats.set(new_quats, offset=(0, 0), fromrank=0)

        return

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "shared": [
                self.elevation
            ]
        }
        if self.boresight_azel is not None:
            req["shared"].append(self.boresight_azel)
        if self.boresight_radec is not None:
            req["shared"].append(self.boresight_radec)
        return req

    def _provides(self):
        return {
            "shared": [
                self.corotator_angle,
            ]
        }

    def _accelerators(self):
        return list()



# Copyright (c) 2022-2025 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import os
import re

import numpy as np
import traitlets
from astropy import units as u

from toast.timing import function_timer, Timer
from toast.traits import trait_docs, Int, Unicode, Bool, Quantity, Float, Instance
from toast.ops.operator import Operator
from toast.utils import Environment, Logger
from toast.observation import default_values as defaults

from ..io import write_book
from ...sim_hardware import telescope_tube_wafer
from ...core.hardware import build_readout_id
from ..instrument import SOFocalplane, update_creation_time


@trait_docs
class SaveBooks(Operator):
    """Export observations to Level-3 Book format.

    Create one book per tube, per observing session.  Each session consists of
    observations from the same telescope tube.  Basic metadata is written to
    Observation frames at the start of each Primary File Group and the
    M_observation.yaml file.

    """

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    book_dir = Unicode("books", help="Top-level export directory")

    focalplane_dir = Unicode(
        None, allow_none=True, help="Directory for focalplane models"
    )

    noise_dir = Unicode(None, allow_none=True, help="Directory for noise models")

    times = Unicode(defaults.times, help="Observation shared key for timestamps")

    shared_flags = Unicode(
        defaults.shared_flags,
        allow_none=True,
        help="Observation shared key for common flags",
    )

    det_data = Unicode(
        defaults.det_data,
        allow_none=True,
        help="Observation detdata key for detector signal",
    )

    det_flags = Unicode(
        defaults.det_flags,
        allow_none=True,
        help="Observation detdata key for detector flags",
    )

    boresight_azel = Unicode(
        defaults.boresight_azel, help="Observation shared key for boresight Az/El"
    )

    boresight_radec = Unicode(
        defaults.boresight_radec,
        allow_none=True,
        help="Observation shared key for boresight RA/DEC",
    )

    corotator_angle = Unicode(
        defaults.corotator_angle,
        allow_none=True,
        help="Observation shared key for corotator_angle (if it is used)",
    )

    boresight_angle = Unicode(
        defaults.boresight_angle,
        allow_none=True,
        help="Observation shared key for boresight rotation angle (if it is used)",
    )

    hwp_angle = Unicode(
        defaults.hwp_angle, allow_none=True, help="Observation shared key for HWP angle"
    )

    frame_intervals = Unicode(
        None,
        allow_none=True,
        help="Observation interval key for frame boundaries",
    )

    creation_time = Float(None, allow_none=True, help="Last detector tuning time")

    gzip = Bool(False, help="If True, gzip compress the frame files")

    purge = Bool(False, help="If True, delete observation data as it is saved")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()
        timer = Timer()
        timer.start()

        if data.comm.ngroups != 1:
            msg = "SaveBooks is only usable in workflows with one process group"
            log.error(msg)
            raise RuntimeError(msg)

        out_fp_dir = self.focalplane_dir
        if self.focalplane_dir is None:
            out_fp_dir = self.book_dir

        out_nse_dir = self.noise_dir
        if self.noise_dir is None:
            out_nse_dir = self.book_dir

        create_time = self.creation_time
        if create_time is None:
            # Find the global minimum time
            for ob in data.obs:
                ob_start = ob.session.start.timestamp()
                if create_time is None:
                    create_time = ob_start
                elif ob_start < create_time:
                    create_time = ob_start

        # Update the focalplane readout_ids to match this creation time
        for ob in data.obs:
            update_creation_time(ob.telescope.focalplane.detector_data, create_time)

        # Split data by session
        data_sessions = data.split(obs_session_name=True)

        fp_cache = dict()

        for sname, sdata in data_sessions.items():
            # Create a full-telescope focalplane once (since it is slow), so that we
            # can reference subsets of detectors from that.
            tele_mat = re.match(r"^(.*?)_", sdata.obs[0].telescope.name)
            telename = tele_mat.group(1)
            if telename in fp_cache:
                full_fp = fp_cache[telename]
            else:
                full_fp = SOFocalplane(
                    sample_rate=sdata.obs[0].telescope.focalplane.sample_rate,
                    telescope=telename,
                    creation_time=create_time,
                )
                fp_cache[telename] = full_fp

            # Pass through all observations and:
            # - group observations by optics tube
            # - ensure that frame boundaries are set up

            tube_obs = dict()

            for ob in sdata.obs:
                tbs = set(ob.telescope.focalplane.detector_data["tube_slot"])
                if len(tbs) != 1:
                    msg = f"Observation {ob.name} has multiple optics tubes ({tbs})"
                    raise RuntimeError(msg)
                this_tube = tbs.pop()
                if this_tube not in tube_obs:
                    tube_obs[this_tube] = list()
                tube_obs[this_tube].append(ob)

                # Create frame intervals if not specified
                redist_sampsets = False
                frame_intervals = self.frame_intervals
                if frame_intervals is None:
                    # We are using the sample set distribution for our frame boundaries.
                    frame_intervals = "frames"
                    timespans = list()
                    offset = 0
                    n_frames = 0
                    first_set = ob.dist.samp_sets[ob.comm.group_rank].offset
                    n_set = ob.dist.samp_sets[ob.comm.group_rank].n_elem
                    for sset in range(first_set, first_set + n_set):
                        for chunk in ob.dist.sample_sets[sset]:
                            first = offset
                            last = min(offset + chunk, ob.n_local_samples - 1)
                            timespans.append(
                                (
                                    ob.shared[self.times][first],
                                    ob.shared[self.times][last],
                                )
                            )
                            n_frames += 1
                            offset += chunk
                    ob.intervals.create_col(
                        frame_intervals, timespans, ob.shared[self.times]
                    )
                else:
                    # We were given an existing set of frame boundaries.  Compute new
                    # sample sets to use when redistributing.
                    if ob.comm_col_rank == 0:
                        # First row of process grid gets local chunks
                        local_sets = list()
                        offset = 0
                        for intr in ob.intervals[frame_intervals]:
                            chunk = intr.last - offset
                            local_sets.append([chunk])
                            offset += chunk
                        if offset != ob.n_local_samples:
                            local_sets.append([ob.n_local_samples - offset])
                        # Gather across the row
                        all_sets = [
                            local_sets,
                        ]
                        if ob.comm_row is not None:
                            all_sets = ob.comm_row.gather(local_sets, root=0)
                        if ob.comm_row_rank == 0:
                            redist_sampsets = list()
                            for pset in all_sets:
                                redist_sampsets.extend(pset)
                    if ob.comm.comm_group is not None:
                        redist_sampsets = ob.comm.comm_group.bcast(
                            redist_sampsets, root=0
                        )

            # Write observations for each tube

            for tube, tobs in tube_obs.items():
                log.info_rank(
                    f"Writing book for session {sname}, tube {tube}",
                    comm=sdata.comm.comm_group,
                )
                # Save it
                write_book(
                    tobs,
                    full_fp,
                    self.book_dir,
                    out_fp_dir,
                    out_nse_dir,
                    self.times,
                    self.shared_flags,
                    self.det_data,
                    self.det_flags,
                    self.boresight_azel,
                    self.boresight_radec,
                    self.corotator_angle,
                    self.boresight_angle,
                    self.hwp_angle,
                    gzip=self.gzip,
                    frame_intervals=frame_intervals,
                    redist_sampsets=redist_sampsets,
                )
                if ob.comm.comm_group is not None:
                    ob.comm.comm_group.barrier()

            # Delete our temporary frame interval if we created it
            for ob in sdata.obs:
                if self.frame_intervals is None:
                    del ob.intervals[frame_intervals]

                if ob.comm.comm_group is not None:
                    ob.comm.comm_group.barrier()

                if self.purge:
                    ob.clear()

        del data_sessions
        if self.purge:
            data.obs.clear()

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "shared": [
                self.times,
                self.boresight_azel,
            ],
            "detdata": [self.det_data],
        }
        if self.boresight_radec is not None:
            req["shared"].append(self.boresight_radec)
        if self.boresight_angle is not None:
            req["shared"].append(self.boresight_angle)
        if self.corotator_angle is not None:
            req["shared"].append(self.corotator_angle)
        if self.hwp_angle is not None:
            req["shared"].append(self.hwp_angle)
        return req

    def _provides(self):
        return dict()



# Copyright (c) 2018-2019 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Focalplane simulation tools.
"""

import re

from collections import OrderedDict

from copy import deepcopy

import numpy as np

from .core import Hardware


def sim_detectors_toast(hw, tele, tube_slots=None):
    """Update hardware model with simulated detector positions.

    Given a Hardware model, generate all detector properties for the specified
    telescope and optionally a subset of optics tube slots (for the LAT).  The
    detector dictionary of the hardware model is updated in place.

    This function requires the toast subpackage (and hence toast) to be
    importable.

    Args:
        hw (Hardware): The hardware object to update.
        tele (str): The telescope name.
        tube_slots (list, optional): The optional list of tube slots to include.

    Returns:
        None

    """
    try:
        from . import toast as sotoast
    except ImportError:
        msg = "Toast package is not importable, cannot simulate detector positions"
        raise RuntimeError(msg)

    sotoast.sim_focalplane.sim_telescope_detectors(
        hw, tele, tube_slots=tube_slots,
    )


def sim_detectors_physical_optics(hw, tele, tube_slots=None):
    """Update hardware model with simulated detector positions.

    Given a Hardware model, generate all detector properties for the specified
    telescope and optionally a subset of optics tube slots (for the LAT).  The
    detector dictionary of the hardware model is updated in place.

    This function uses information from physical optics simulations to estimate
    the location of detectors.

    Args:
        hw (Hardware): The hardware object to update.
        tele (str): The telescope name.
        tube_slots (list, optional): The optional list of tube slots to include.

    Returns:
        None

    """
    raise NotImplementedError("Not yet implemented")


def sim_nominal():
    """Return a simulated nominal hardware configuration.

    This returns a simulated Hardware object with the nominal instrument
    properties / metadata, but with an empty set of detector locations.

    This can then be passed to one of the detector simulation functions
    to build up the list of detectors.

    Returns:
        (Hardware): Hardware object with nominal metadata.

    """
    cnf = OrderedDict()

    bands = OrderedDict()

    bnd = OrderedDict()
    bnd["center"] = 25.7
    bnd["low"] = 21.7
    bnd["high"] = 29.7
    bnd["bandpass"] = ""
    bnd["NET"] = 435.1
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    # Noise elevation scaling fits from Carlos Sierra
    # These numbers are for V3 LAT baseline
    bnd["A"] = 0.06
    bnd["C"] = 0.92
    bnd["NET_corr"] = 1.10
    bands["LAT_f030"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 38.9
    bnd["low"] = 30.9
    bnd["high"] = 46.9
    bnd["bandpass"] = ""
    bnd["NET"] = 281.5
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = 0.16
    bnd["C"] = 0.79
    bnd["NET_corr"] = 1.02
    bands["LAT_f040"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 92.0
    bnd["low"] = 79.0
    bnd["high"] = 105.0
    bnd["bandpass"] = ""
    bnd["NET"] = 361.0
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = 0.16
    bnd["C"] = 0.80
    bnd["NET_corr"] = 1.09
    bands["LAT_f090"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 147.5
    bnd["low"] = 130.0
    bnd["high"] = 165.0
    bnd["bandpass"] = ""
    bnd["NET"] = 352.4
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = 0.17
    bnd["C"] = 0.78
    bnd["NET_corr"] = 1.01
    bands["LAT_f150"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 225.7
    bnd["low"] = 196.7
    bnd["high"] = 254.7
    bnd["bandpass"] = ""
    bnd["NET"] = 724.4
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = 0.29
    bnd["C"] = 0.62
    bnd["NET_corr"] = 1.02
    bands["LAT_f230"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 285.4
    bnd["low"] = 258.4
    bnd["high"] = 312.4
    bnd["bandpass"] = ""
    bnd["NET"] = 1803.9
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = 0.36
    bnd["C"] = 0.53
    bnd["NET_corr"] = 1.00
    bands["LAT_f290"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 25.7
    bnd["low"] = 21.7
    bnd["high"] = 29.7
    bnd["bandpass"] = ""
    bnd["NET"] = 314.1
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    # Noise elevation scaling fits from Carlos Sierra
    # These numbers are for V3 SAT baseline
    bnd["A"] = 0.06
    bnd["C"] = 0.92
    bnd["NET_corr"] = 1.06
    bands["SAT_f030"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 38.9
    bnd["low"] = 30.9
    bnd["high"] = 46.9
    bnd["bandpass"] = ""
    bnd["NET"] = 225.8
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = 0.19
    bnd["C"] = 0.76
    bnd["NET_corr"] = 1.01
    bands["SAT_f040"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 92.0
    bnd["low"] = 79.0
    bnd["high"] = 105.0
    bnd["bandpass"] = ""
    bnd["NET"] = 245.1
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = 0.19
    bnd["C"] = 0.76
    bnd["NET_corr"] = 1.04
    bands["SAT_f090"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 147.5
    bnd["low"] = 130.0
    bnd["high"] = 165.0
    bnd["bandpass"] = ""
    bnd["NET"] = 250.2
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = 0.23
    bnd["C"] = 0.70
    bnd["NET_corr"] = 1.02
    bands["SAT_f150"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 225.7
    bnd["low"] = 196.7
    bnd["high"] = 254.7
    bnd["bandpass"] = ""
    bnd["NET"] = 540.3
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = 0.35
    bnd["C"] = 0.54
    bnd["NET_corr"] = 1.00
    bands["SAT_f230"] = bnd

    bnd = OrderedDict()
    bnd["center"] = 285.4
    bnd["low"] = 258.4
    bnd["high"] = 312.4
    bnd["bandpass"] = ""
    bnd["NET"] = 1397.5
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = 0.42
    bnd["C"] = 0.45
    bnd["NET_corr"] = 1.00
    bands["SAT_f290"] = bnd

    # Special "band" for dark bolometers

    bnd = OrderedDict()
    bnd["center"] = np.nan
    bnd["low"] = np.nan
    bnd["high"] = np.nan
    bnd["bandpass"] = ""
    bnd["NET"] = 1000.0
    bnd["fknee"] = 50.0
    bnd["fmin"] = 0.01
    bnd["alpha"] = 1.0
    bnd["A"] = np.nan
    bnd["C"] = np.nan
    bnd["NET_corr"] = 1.00
    bands["NC"] = bnd

    cnf["bands"] = bands

    wafer_slots = OrderedDict()

    wtypes = ["LAT_UHF", "SAT_UHF", "LAT_MF", "SAT_MF", "LAT_LF", "SAT_LF"]
    wcnt = {
        "LAT_LF": 1*3,
        "LAT_MF": 4*3,
        "LAT_UHF": 2*3,
        "SAT_LF": 1*7,
        "SAT_MF": 2*7,
        "SAT_UHF": 1*7
    }
    wnp = {
        "LAT_LF": 37,
        "LAT_MF": 432,
        "LAT_UHF": 432,
        "SAT_LF": 37,
        "SAT_MF": 432,
        "SAT_UHF": 432
    }
    wpixmm = {
        "LAT_LF": 18.0,
        "LAT_MF": 5.3,
        "LAT_UHF": 5.3,
        "SAT_LF": 18.0,
        "SAT_MF": 5.3,
        "SAT_UHF": 5.3
    }
    wrhombgap = {
        "LAT_MF": 0.71,
        "LAT_UHF": 0.71,
        "SAT_MF": 0.71,
        "SAT_UHF": 0.71
    }
    wbd = {
        "LAT_LF": ["LAT_f030", "LAT_f040"],
        "LAT_MF": ["LAT_f090", "LAT_f150"],
        "LAT_UHF": ["LAT_f230", "LAT_f290"],
        "SAT_LF": ["SAT_f030", "SAT_f040"],
        "SAT_MF": ["SAT_f090", "SAT_f150"],
        "SAT_UHF": ["SAT_f230", "SAT_f290"]
    }
    windx = 0
    cardindx = 0
    for wt in wtypes:
        for ct in range(wcnt[wt]):
            wn = "w{:02d}".format(windx)
            wf = OrderedDict()
            wf["type"] = wt
            if ((wt == "LAT_LF") or (wt == "SAT_LF")):
                wf["packing"] = "S"
            else:
                wf["packing"] = "F"
                wf["rhombusgap"] = wrhombgap[wt]
            wf["npixel"] = wnp[wt]
            wf["pixsize"] = wpixmm[wt]
            wf["bands"] = wbd[wt]
            wf["card_slot"] = "card_slot{:02d}".format(cardindx)
            wf["wafer_name"] = ""
            cardindx += 1
            wafer_slots[wn] = wf
            windx += 1

    cnf["wafer_slots"] = wafer_slots

    tube_slots = OrderedDict()

    woff = {
        "LAT_LF": 0,
        "LAT_MF": 0,
        "LAT_UHF": 0,
        "SAT_LF": 0,
        "SAT_MF": 0,
        "SAT_UHF": 0
    }

    ltubes = ["LAT_UHF", "LAT_UHF", "LAT_MF", "LAT_MF", "LAT_MF", "LAT_MF", "LAT_LF"]

    # The optics tubes are arranged using several conventions and here we map between
    # them.  Note that for this hardware model, the projection on the sky is defined
    # at 60 degree elevation with no boresight rotation.

    # TOAST hexagon layout positions in Xi/Eta coordinates
    ltube_toasthex_pos = [0, 1, 2, 3, 5, 6, 10]

    # "Optics" locations as given in several SO slide decks.  Note this is flipped in
    # some figures.  See:
    # https://simonsobs.atlassian.net/wiki/spaces/PRO/pages/101974017/Focal+Plane+Coordinates
    ltube_optics_pos = [1, 3, 5, 4, 8, 9, 12]

    # Cryo team names for these positions
    ltube_cryonames=["c1", "i5", "i6", "i1", "i3", "i4", "o6"]

    lat_ufm_slot = [
        0,
        1,
        2,
    ]

    lat_ufm_thetarot = [
        240.0,
        0.0,
        120.0,
    ]

    for tindx in range(7):
        nm = ltube_cryonames[tindx]
        ttyp = ltubes[tindx]
        tb = OrderedDict()
        tb["type"] = ttyp
        tb["waferspace"] = 128.4
        tb["wafer_slots"] = list()
        tb["wafer_slot_angle"] = [
            lat_ufm_thetarot[tw] for tw in range(3)
        ] # Degrees
        # The "slot" here is the relative slot (ws0 - ws2) within the tube.
        tb["wafer_ufm_slot"] = list()
        for tw in range(3):
            off = 0
            for w, props in cnf["wafer_slots"].items():
                if props["type"] == ttyp:
                    if off == woff[ttyp]:
                        props["tube_index"] = tw
                        tb["wafer_slots"].append(w)
                        tb["wafer_ufm_slot"].append(lat_ufm_slot[tw])
                        woff[ttyp] += 1
                        break
                    off += 1
        tb["toast_hex_pos"] = ltube_toasthex_pos[tindx]
        tb["optics_pos"] = ltube_optics_pos[tindx]
        tb["tube_name"] = ""
        tb["receiver_name"] = ""
        tube_slots[nm] = tb

    # These are taken from:
    # https://simonsobs.atlassian.net/wiki/spaces/PRO/pages/101974017/Focal+Plane+Coordinates#UFM-Layout.1
    hex_to_ufm_slot = [
        0,
        2,
        1,
        6,
        5,
        4,
        3,
    ]
    hex_to_ufm_loc = [
        "AX",
        "NE",
        "NO",
        "NW",
        "SW",
        "SO",
        "SE",
    ]
    hex_to_ufm_thetarot = [
        240.0,
        0.0,
        300.0,
        180.0,
        120.0,
        60.0,
        60.0,
    ]

    stubes = ["SAT_MF", "SAT_MF", "SAT_UHF", "SAT_LF"]
    for tindx in range(4):
        nm = "ST{:d}".format(tindx+1)
        ttyp = stubes[tindx]
        tb = OrderedDict()
        tb["type"] = ttyp
        tb["waferspace"] = 128.4
        tb["wafer_slots"] = list()
        tb["wafer_slot_angle"] = [
            hex_to_ufm_thetarot[tw] for tw in range(7)
        ] # Degrees
        # The "slot" here is the relative slot (ws0 - ws6) within the tube.
        tb["wafer_ufm_slot"] = list()
        # The "loc" here is the compass direction name (e.g. NO, NE, SW, etc.)
        tb["wafer_ufm_loc"] = list()
        for tw in range(7):
            off = 0
            for w, props in cnf["wafer_slots"].items():
                if props["type"] == ttyp:
                    if off == woff[ttyp]:
                        props["tube_index"] = tw
                        tb["wafer_slots"].append(w)
                        tb["wafer_ufm_slot"].append(hex_to_ufm_slot[tw])
                        tb["wafer_ufm_loc"].append(hex_to_ufm_loc[tw])
                        woff[ttyp] += 1
                        break
                    off += 1
        tb["toast_hex_pos"] = 0
        tb["optics_pos"] = 0
        tb["tube_name"] = ""
        tb["receiver_name"] = ""
        tube_slots[nm] = tb

    cnf["tube_slots"] = tube_slots

    telescopes = OrderedDict()

    tele = OrderedDict()
    tele["tube_slots"] = ["c1", "i5", "i6", "i1", "i3", "i4", "o6"]
    tele["platescale"] = 0.00495
    # This tube spacing in mm corresponds to 1.78 degrees projected on
    # the sky at a plate scale of 0.00495 deg/mm.
    tele["tubespace"] = 359.6
    fwhm = OrderedDict()
    fwhm["LAT_f030"] = 7.4
    fwhm["LAT_f040"] = 5.1
    fwhm["LAT_f090"] = 2.2
    fwhm["LAT_f150"] = 1.4
    fwhm["LAT_f230"] = 1.0
    fwhm["LAT_f290"] = 0.9
    tele["fwhm"] = fwhm
    tele["platform_name"] = ""
    telescopes["LAT"] = tele

    fwhm_sat = OrderedDict()
    fwhm_sat["SAT_f030"] = 91.0
    fwhm_sat["SAT_f040"] = 63.0
    fwhm_sat["SAT_f090"] = 30.0
    fwhm_sat["SAT_f150"] = 17.0
    fwhm_sat["SAT_f230"] = 11.0
    fwhm_sat["SAT_f290"] = 9.0

    tele = OrderedDict()
    tele["tube_slots"] = ["ST1"]
    tele["platescale"] = 0.09668
    tele["fwhm"] = fwhm_sat
    tele["platform_name"] = ""
    telescopes["SAT1"] = tele

    tele = OrderedDict()
    tele["tube_slots"] = ["ST2"]
    tele["platescale"] = 0.09668
    tele["fwhm"] = fwhm_sat
    tele["platform_name"] = ""
    telescopes["SAT2"] = tele

    tele = OrderedDict()
    tele["tube_slots"] = ["ST3"]
    tele["platescale"] = 0.09668
    tele["fwhm"] = fwhm_sat
    tele["platform_name"] = ""
    telescopes["SAT3"] = tele

    tele = OrderedDict()
    tele["tube_slots"] = ["ST4"]
    tele["platescale"] = 0.09668
    tele["fwhm"] = fwhm_sat
    tele["platform_name"] = ""
    telescopes["SAT4"] = tele

    cnf["telescopes"] = telescopes

    card_slots = OrderedDict()
    crate_slots = OrderedDict()

    crt_indx = 0

    for tel in cnf["telescopes"]:
        crn = "crate_slot{:02d}".format(crt_indx)
        crt = OrderedDict()
        crt["card_slots"] = list()
        crt["telescope"] = tel
        crt["crate_name"] = ""

        ## get all the wafer card numbers for a telescope
        tb_wfrs = [cnf["tube_slots"][t]["wafer_slots"] for t in cnf["telescopes"][tel]["tube_slots"]]
        tl_wfrs = [i for sl in tb_wfrs for i in sl]
        wafer_cards = [cnf["wafer_slots"][w]["card_slot"] for w in tl_wfrs]

        # add all cards to the card table and assign to crates
        for crd in wafer_cards:
            cdprops = OrderedDict()
            cdprops["nbias"] = 12
            cdprops["nAMC"] = 2
            cdprops["nchannel"] = 1764
            cdprops["card_name"] = ""
            card_slots[crd] = cdprops

            crt["card_slots"].append(crd)

            # name new crates when current one is full
            if ('S' in tel and len(crt["card_slots"]) >=4) or len(crt["card_slots"]) >=6:
                crate_slots[crn] = crt
                crt_indx += 1
                crn = "crate_slot{:02d}".format(crt_indx)
                crt = OrderedDict()
                crt["card_slots"] = list()
                crt["telescope"] = tel
                crt["crate_name"] = ""

        # each telescope starts with a new crate
        crate_slots[crn] = crt
        crt_indx += 1

    cnf["card_slots"] = card_slots
    cnf["crate_slots"] = crate_slots

    # Add an empty set of detectors here, in case the user just wants access to
    # the hardware metadata.
    cnf["detectors"] = OrderedDict()

    hw = Hardware()
    hw.data = cnf

    return hw


def telescope_tube_wafer():
    """Global mapping of telescopes, tubes, and wafers used in simulations.

    This mapping is here rather than core.hardware, so that we could put
    alternate definitions there for actual fielded configurations.

    Returns:
        (dict):  The mapping

    """
    hw = sim_nominal()
    result = dict()
    for tele_name, tele_props in hw.data["telescopes"].items():
        tb = dict()
        for tube_name in tele_props["tube_slots"]:
            tube_props = hw.data["tube_slots"][tube_name]
            tb[tube_name] = list(tube_props["wafer_slots"])
        result[tele_name] = tb
    return result



# Copyright (c) 2020-2021 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simons Observatory simulation and processing modules using TOAST.

"""

# On first import, set toast default names to use across all built-in operators.

import toast

toast.observation.set_default_values(
    {
        "times": "times",
        "shared_flags": "flags",
        "det_data": "signal",
        "det_flags": "flags",
        "hwp_angle": "hwp_angle",
        "azimuth": "azimuth",
        "elevation": "elevation",
        "roll": "roll",
        "boresight_azel": "boresight_azel",
        "boresight_radec": "boresight_radec",
        "position": "position",
        "velocity": "velocity",
        "pixels": "pixels",
        "weights": "weights",
        "quats": "quats",
        "quats_azel": "quats_azel",
        "corotator_angle": "corotator_angle",
        "boresight_angle": "boresight_angle",
    }
)

from .instrument import SOFocalplane, simulated_telescope
from .sim_focalplane import sim_telescope_detectors



# Copyright (c) 2018-2020 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simons Observatory script entry points.

This module contains files which have a main() function defined, and which are
defined as entry points in setup.py.

"""



# Copyright (c) 2020 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check functionality of some coordinate routines.

"""

import itertools
import unittest
import time
import numpy as np

from sotodlib import coords, core
from sotodlib.coords import optics as co
import so3g
from pixell import enmap

DEG = np.pi/180
CRVAL = [34.0*DEG, -20.9*DEG]
TOL_RAD = .00001 * DEG

def get_sightline():
    ra0, dec0, gamma0 = CRVAL[0], CRVAL[1], 0
    ra, dec, gamma = [np.array([_x]) for _x in [ra0, dec0, gamma0]]
    sight = so3g.proj.CelestialSightLine()
    sight.Q = (
        so3g.proj.quat.euler(2, ra) *
        so3g.proj.quat.euler(1, np.pi/2 - dec) *
        so3g.proj.quat.euler(2, gamma)
    )
    return sight


class FP:
    def __init__(self, xi, eta):
        self.xi = np.array(xi)
        self.eta = np.array(eta)


class FootprintTest(unittest.TestCase):
    def test_10_kernels(self):
        """Test creation of sensible WCS kernels for basic cases."""
        ra0, dec0 = CRVAL
        res = 0.01 * DEG

        # Test zenithal -- (ra0, dec0) is the reference point.
        for proj in ['TAN', 'ZEA']:
            wcsk = coords.get_wcs_kernel(proj, ra0, dec0, res)
            msg = f'Check crpix for {proj}'
            self.assertAlmostEqual(wcsk.wcs.crpix[0], 1, delta=TOL_RAD, msg=msg)
            self.assertAlmostEqual(wcsk.wcs.crpix[1], 1, delta=TOL_RAD, msg=msg)

        # Test cylindrical -- pixell puts the crval[1] on the equator
        # and dec0 is used for the conformal latitude.
        for proj in ['CAR', 'CEA']:
            wcsk = coords.get_wcs_kernel(proj, ra0, dec0, res)
            msg = f'Check crpix for {proj}'
            self.assertAlmostEqual(wcsk.wcs.crpix[0], 1, delta=TOL_RAD, msg=msg)
            self.assertNotAlmostEqual(wcsk.wcs.crpix[1], 1, delta=TOL_RAD, msg=msg)

        # This is going to break.
        fp = FP(xi =[0., -0.01*DEG],
                eta=[0., -0.01*DEG])
        sight = get_sightline()
        tod = core.AxisManager(core.LabelAxis('dets', ['a']))
        fp = coords.get_footprint(tod, wcs_kernel=wcsk, focal_plane=fp, sight=sight)

    def test_20_supergeom_simple(self):
        """Check that coords.get_supergeom does sensible things in simple cases."""
        for proj in ['TAN', 'CEA']:
            ra0, dec0 = CRVAL
            res = 0.01 * DEG
            wcs = coords.get_wcs_kernel(proj, ra0, dec0, res)

            wcs.wcs.crpix = (60, 70)
            map0 = enmap.zeros((100,200), wcs=wcs)
            map0[2, 3] = 10.
            map0[90, 192] = 11.

            # Extracts.
            m1 = map0[:10,:10]
            m2 = map0[-10:,-10:]
            
            # Reconstruct.
            sg = coords.get_supergeom((m1.shape, m1.wcs), (m2.shape, m2.wcs))
            mapx = enmap.zeros(*sg)
            mapx.insert(m1)
            mapx.insert(m2)
            self.assertTupleEqual(map0.shape, mapx.shape)
            self.assertTrue(np.all(mapx==map0))

    def test_30_supergeom_translate(self):
        """Check that coords.get_supergeom does sensible thing for maps in
        cylindrical projections with compatible but not identical
        crval.

        """
        proj = 'CAR'
        ra0, dec0 = CRVAL
        res = 0.01 * DEG
        wcs = coords.get_wcs_kernel(proj, ra0, dec0, res)

        wcs.wcs.crpix = (60, 70)
        map0 = enmap.zeros((100,200), wcs=wcs)
        map0[2, 3] = 10.
        map0[90, 192] = 11.

        # Extracts.
        m1 = map0[:10,:10]
        m2 = map0[-10:,-10:]

        # In simple cylindrical projections, there's a degeneracy
        # between crval and crpix in the longitude component -- crval
        # can be anywhere on the equator.  It is useful to be able to
        # join maps even if they have different crval[0], provided the
        # pixel centers line up.  (The same is not true of crval[1],
        # which tips the native equator relative to the celestial
        # equator.)

        for axis, should_work in [(0, True), (1, False)]:
            dpix = 10.5
            m2 = map0[-10:,-10:]
            m2.wcs.wcs.crpix[axis] += dpix
            m2.wcs.wcs.crval[axis] += dpix * m2.wcs.wcs.cdelt[axis]

            if should_work:
                sg = coords.get_supergeom((m1.shape, m1.wcs), (m2.shape, m2.wcs))
                mapx = enmap.zeros(*sg)
                mapx.insert(m1)
                mapx.insert(m2)
                self.assertTupleEqual(map0.shape, mapx.shape,
                                      msg="Reconstructed map shape.")
                self.assertTrue(np.all(mapx==map0),
                                msg="Reconstructed map data.")

            else:
                msg = "Translating crval in dec should cause "\
                    "coord consistency check failure."
                with self.assertRaises(ValueError, msg=msg):
                    sg = coords.get_supergeom((m1.shape, m1.wcs), (m2.shape, m2.wcs))

class CoordsUtilsTest(unittest.TestCase):
    def test_valid_arg(self):
        from sotodlib.coords.helpers import _valid_arg
        tod = core.AxisManager()
        tod.wrap('a', np.array([1,2,3]))
        self.assertIs(_valid_arg(None, 'a', src=tod), tod.a)
        self.assertIs(_valid_arg('a', None, src=tod), tod.a)
        self.assertIs(_valid_arg(None, tod.a), tod.a)
        self.assertIs(_valid_arg(tod.get('b')), None)
        self.assertIs(_valid_arg(tod.get('b'), 'a', src=tod), tod.a)

    def test_scalar_last_quat(self):
        test_array = np.array([[2,3,4,1],[90,100,23,14]])

        # Convert one quat
        qa = coords.ScalarLastQuat(test_array[0])
        self.assertIsInstance(qa, np.ndarray)
        q3 = qa.to_g3()
        self.assertIsInstance(q3, so3g.proj.quat.quat)
        self.assertEqual(q3.a, 1)
        qb = coords.ScalarLastQuat(q3)
        np.testing.assert_array_equal(qa, qb)

        # Convert a vector of quats
        qa = coords.ScalarLastQuat(test_array)
        v3 = qa.to_g3()
        self.assertIsInstance(v3, so3g.proj.quat.G3VectorQuat)
        self.assertEqual(v3[0].a, 1)
        self.assertEqual(v3[1].a, 14)
        qb = coords.ScalarLastQuat(v3)
        np.testing.assert_array_equal(qa, qb)

    def test_cover(self):
        x0, y0 = 1*DEG, 4*DEG
        R = 5*DEG
        x = np.linspace(-R, R, 50)
        y = np.linspace(-R, R, 45)
        xy = np.transpose(list(itertools.product(x, y)))
        s = xy[0]**2 + xy[1]**2 < R**2
        
        xy = xy[:,s] + np.array([x0, y0])[:,None]
        (xi0, eta0), R0, (xi, eta) = \
            coords.helpers.get_focal_plane_cover(count=16, xieta=xy)
        np.testing.assert_allclose([xi0, eta0, R], [x0, y0, R0],
                                   atol=R*0.05)
        self.assertEqual(len(xi), 16)

        # Works with nans?
        xy[0,0] = np.nan
        coords.helpers.get_focal_plane_cover(xieta=xy)

        # Exclude dets using det_weights?
        det_weights = np.ones(xy.shape[1])
        det_weights[3:34] = 0.
        for dtype in ['float', 'int', 'bool']:
            coords.helpers.get_focal_plane_cover(
                xieta=xy, det_weights=det_weights.astype(dtype))

        # Works for only a single det?
        det_weights[2:] = 0.
        (xi0, eta0), R0, _ = \
            coords.helpers.get_focal_plane_cover(xieta=xy, det_weights=det_weights)

        # Fails if all dets excluded somehow?
        det_weights[1] = 0.
        with self.assertRaises(ValueError):
            coords.helpers.get_focal_plane_cover(xieta=xy, det_weights=det_weights)

        # Fails with all nans?
        xy[1,1:] = np.nan
        with self.assertRaises(ValueError):
            coords.helpers.get_focal_plane_cover(xieta=xy)

    def test_source_pos_ephem(self):
        """
        Test getting the position of a Solar System object.
        """
        t = 1.7e9
        pos1 = coords.planets.get_source_pos('jupiter', t)
        pos1 = coords.planets.get_source_azel('jupiter', t)

    def test_source_pos_fixed(self):
        """
        Test getting the position of a fixed source.
        """
        t = time.time()

        pos1 = coords.planets.get_source_pos("J1000+1000", t)
        pos2 = coords.planets.get_source_pos("j1000p1000", t)
        self.assertEqual(pos1[0], pos2[0])
        self.assertEqual(pos1[1], pos2[1])

        pos1 = coords.planets.get_source_pos("J1000-1000", t)
        pos2 = coords.planets.get_source_pos("j1000m1000", t)
        self.assertEqual(pos1[0], pos2[0])
        self.assertEqual(pos1[1], pos2[1])

        # This doesn't work but it probably should...
        #coords.planets.get_source_azel("J1000-1000", t)
            

class OpticsTest(unittest.TestCase):
    def test_sat_fp(self):
        x = np.array([-100, 0, 100]) 
        y = x.copy()
        pol = x.copy()

        xi, eta, gamma = co.get_focal_plane(None, x, y, pol, 0, "SAT", "ws1", ufm_to_fp_pars={'theta': 60.0, 'dx': 0.0, 'dy': 128.5})
        self.assertTrue(np.all(np.isclose(xi, np.array([-6.4406e-02,  0,  5.58489e-02]))))
        self.assertTrue(np.all(np.isclose(eta, np.array([0.01425728, -0.2207397, -0.404499]))))
        self.assertTrue(np.all(np.isclose(gamma, np.array([5.409, 3.6846, 1.8156]))))

if __name__ == '__main__':
    unittest.main()



import sqlite3
import os
import sys
import argparse
from collections import OrderedDict
import numpy as np

from . import common
from .. import util
from .resultset import ResultSet

_DB_VERSION = 3

TABLE_DEFS = {
    'detsets': [
        "`name`    varchar(16)",
        "`det`     varchar(32)",
        "CONSTRAINT name_det UNIQUE (name, det)",
    ],
    'files': [
        "`name`    varchar(256) unique",
        "`detset`  varchar(16)",
        "`obs_id`  varchar(256)",
        "`sample_start` int",
        "`sample_stop`  int",
    ],
    'frame_offsets': [
        "`file_name` varchar(256)",
        "`frame_index` int",
        "`byte_offset` int",
        "`frame_type` varchar(16)",
        "`sample_start` int",
        "`sample_stop` int",
    ],
    'meta': [
        "`param` varchar(32) UNIQUE",
        "`value` varchar",
    ],
    '_indices': [
        "CREATE INDEX IF NOT EXISTS idx_obs_id_column ON files(obs_id)",
        "CREATE INDEX IF NOT EXISTS idx_detset_column ON detsets(name)",
    ],
}


class ObsFileDb:
    """sqlite3-based database for managing large archives of files.

    The data model here is that each distinct "Observation" comprises
    co-sampled detector data for a large number of detectors.  Each
    detector belongs to a single "detset", and there is a set of files
    containing the data for each detset.  Finding the file that
    contains data for a particular detector is a matter of looking up
    what detset the detector is in, and looking up what file covers
    that detset.

    Note that many functions have a "commit" option, which simply
    affects whether the .commit is called on the database or not (it
    can be faster to suppress commit ops when a running a batch of
    updates, and commit manually at the end).

    """

    #: The sqlite3 database connection.
    conn = None

    #: Path relative to which filenames in the database should be
    #: interpreted.  This only applies to relative filenames (those not
    #: starting with /).
    prefix = ''

    def __init__(self, map_file=None, prefix=None, init_db=True, readonly=False):
        """Instantiate an ObsFileDb.

        Arguments:
          map_file (string): sqlite database file to map.  Defaults to
            ':memory:'.
          prefix (string): as described in class documentation.
          init_db (bool): If True, attempt to create the database
            tables.
          readonly (bool): If True, the database file will be mapped
            in read-only mode.  Not valid on dbs held in :memory:.

        """
        if isinstance(map_file, sqlite3.Connection):
            self.conn = map_file
        else:
            if map_file is None:
                map_file = ':memory:'
            self.conn = sqlite3.connect(map_file)
            uri = False
            if readonly:
                if map_file == ':memory:':
                    raise ValueError('Cannot honor request for readonly db '
                                     'mapped to :memory:.')
                map_file, uri = 'file:%s?mode=ro' % map_file, True

            self.conn = sqlite3.connect(map_file, uri=uri)

        self.conn.row_factory = sqlite3.Row  # access columns by name

        self.prefix = self._get_prefix(map_file, prefix)
        if init_db and not readonly:
            self._create()

    @staticmethod
    def _get_prefix(map_file, prefix):
        """Common logic for setting the file prefix based on map_file and
        prefix arguments.

        """
        if prefix is not None:
            return prefix
        if map_file == ':memory:':
            return ''
        return os.path.split(os.path.abspath(map_file))[0] + '/'

    def to_file(self, filename, overwrite=True, fmt=None):
        """Write the present database to the indicated filename.

        Args:
          filename (str): the path to the output file.
          overwrite (bool): whether an existing file should be
            overwritten.
          fmt (str): 'sqlite', 'dump', or 'gz'.  Defaults to 'sqlite'
            unless the filename ends with '.gz', in which it is 'gz'.

        """
        return common.sqlite_to_file(self.conn, filename, overwrite=overwrite, fmt=fmt)

    @classmethod
    def from_file(cls, filename, prefix=None, fmt=None, force_new_db=True):
        """This method calls
            :func:`sotodlib.core.metadata.common.sqlite_from_file`
        """
        conn = common.sqlite_from_file(filename, fmt=fmt,
                                       force_new_db=force_new_db)
        if prefix is None:
            prefix = os.path.split(filename)[0] + '/'
        return cls(conn, init_db=False, prefix=prefix, )

    @classmethod
    def for_dir(cls, path, filename='obsfiledb.sqlite', readonly=True):
        """Deprecated; use from_file()."""
        print('Use of ObsFileDb.for_dir() is deprecated... use from_file.')
        return cls.from_file(os.path.join(path, filename), prefix=path)

    def copy(self, map_file=None, overwrite=False):
        """
        Duplicate the current database into a new database object, and
        return it.  If map_file is specified, the new database will be
        connected to that sqlite file on disk.  Note that a quick way
        of writing a Db to disk to call copy(map_file=...).
        """
        if map_file is None:
            map_file = ':memory:'
        script = ' '.join(self.conn.iterdump())
        if map_file != ':memory:' and os.path.exists(map_file):
            if not overwrite:
                raise RuntimeError("Output database '%s' exists -- remove or "
                                   "pass overwrite=True to copy." % map_file)
            os.remove(map_file)
        new_db = ObsFileDb(map_file, init_db=False)
        new_db.conn.executescript(script)
        new_db.prefix = self.prefix
        return new_db

    def _get_version(self, conn=None):
        if conn is None:
            conn = self.conn
        rows = conn.execute("select value from meta where "
                            "param='obsfiledb_version'").fetchall()
        if len(rows) == 0:
            return None
        return int(rows[0][0])

    def _create(self):
        """
        Create the database tables if they do not already exist.
        """
        # Create the tables:
        c = self.conn.cursor()
        for table_name, column_defs in TABLE_DEFS.items():
            if table_name.startswith('_'):
                continue
            q = ('create table if not exists `%s` (' % table_name  +
                 ','.join(column_defs) + ')')
            c.execute(q)

        for index in TABLE_DEFS['_indices']:
            c.execute(index)

        if self._get_version(conn=c) is None:
            c.execute('insert or ignore into meta (param,value) values (?,?)',
                      ('obsfiledb_version', _DB_VERSION))

        self.conn.commit()

    def add_detset(self, detset_name, detector_names, commit=True):
        """Add a detset to the detsets table (by adding detectors with
        specific names to it).

        Arguments:
          detset_name (str): The name of the detset.
          detector_names (list of str): New detectors belonging to
            this detset.

        """
        for d in detector_names:
            q = 'insert into detsets (name,det) values (?,?)'
            self.conn.execute(q, (detset_name, d))
        if commit:
            self.conn.commit()

    def add_obsfile(self, filename, obs_id, detset, sample_start=None, sample_stop=None,
                    commit=True):
        """Add an observation file to the files table.

        Arguments:
          filename (str): The filename, relative to the data base
            directory and without a leading /.
          obs_id (str): The observation id.
          detset (str): The detset name.
          sample_start (int): The observation sample index at the
            start of this file.
          sample_stop (int): sample_start + n_samples.

        """
        self.conn.execute(
            'insert into files (name,detset,obs_id,sample_start,sample_stop) '
            'values (?,?,?,?,?)',
            (filename,detset,obs_id,sample_start,sample_stop))
        if commit:
            self.conn.commit()

    # Retrieval

    def get_obs(self):
        """Returns all a list of all obs_id present in this database.

        """
        c = self.conn.execute('select distinct obs_id from files')
        return [r[0] for r in c]
    
    def get_obs_with_detset(self, detset):
        """Returns a list of all obs_ids that include a specified detset"""
        c = self.conn.execute(
            f"select distinct obs_id from files where detset='{detset}'"
        )
        return [r[0] for r in c]

    def get_detsets(self, obs_id):
        """Returns a list of all detsets represented in the observation
        specified by obs_id.

        """
        c = self.conn.execute('select distinct detset from files '
                              'where obs_id=?', (obs_id,))
        return [r[0] for r in c]

    def get_dets(self, detset):
        """Returns a list of all detectors in the specified detset.

        """
        c = self.conn.execute('select det from detsets where name=?', (detset,))
        return [r[0] for r in c]

    def get_det_table(self, obs_id):
        """Get table of detectors and detsets suitable for use with Context
        det_info.  Returns Resultset with keys=['dets:detset','dets:readout_id'].

        """
        c = self.conn.execute(
            'select distinct detsets.name as `dets:detset`, det as `dets:readout_id`'
            'from detsets join files '
            'on files.detset=detsets.name where obs_id=?', (obs_id, ))
        return ResultSet.from_cursor(c)

    def get_files(self, obs_id, detsets=None, prefix=None):
        """Get the file names associated with a particular obs_id and detsets.

        Returns:

          OrderedDict where the key is the detset name and the value
          is a list of tuples of the form (full_filename,
          sample_start, sample_stop).

        """
        if prefix is None:
            prefix = self.prefix

        if detsets is None:
            c = self.conn.execute('select detset, name, sample_start, sample_stop '
                                  'from files where obs_id=? '
                                  'order by detset, sample_start',
                                  (obs_id,))
        else:
            c = self.conn.execute('select detset, name, sample_start, sample_stop '
                                  'from files where obs_id=? and detset in (%s) '
                                  'order by detset, sample_start' %
                                  ','.join(['?' for _ in detsets]),
                                  (obs_id,) + tuple(detsets))
        output = OrderedDict()
        for r in c:
            if not r[0] in output:
                output[r[0]] = []
            output[r[0]].append((os.path.join(prefix, r[1]), r[2], r[3]))
        return output

    def lookup_file(self, filename, resolve_paths=True, prefix=None, fail_ok=False):
        """Determine what, if any, obs_id (and detset and sample range) is
        associated with the specified data file.

        Args:
          filename (str): a string corresponding to a file that is
            covered by this db.  See note on how this is resolved.
          resolve_paths (bool): If True, then the incoming filename is
            treated as a path to a specific file on disk, and the
            database is queried for files that also resolve to that
            are equivalent to that same file on disk (accounting for
            prefix).  If False, then the incoming filename is taken as
            opaque text to match against the corresponding entries in
            the obsfiledb file "name" column (including whatever path
            information is in either of those strings).
          fail_ok (bool): If True, then None is returned if the
            filename is not found in the db (instead of raising
            RuntimeError).

        Returns:
          A dict with entries:
          - ``obs_id``: The obs_id
          - ``detsets``: A list containing the name of the single detset
            covered by this file.
          - ``sample_range``: A tuple with the start and stop sample
            indices for this file.

        """
        if prefix is None:
            prefix = self.prefix

        if resolve_paths:
            # Clarify our target ...
            filename = os.path.realpath(filename)
            basename = os.path.split(filename)[1]
            # Do a non-conclusive match against the basename ...
            c = self.conn.execute(
                'select name, obs_id, detset, sample_start, sample_stop '
                'from files where name like ?', ('%' + basename, ))
            rows = c.fetchall()
            # Keep only the rows that are definitely our target file.
            rows = [r for r in rows
                    if os.path.realpath(os.path.join(prefix, r[0])) == filename]
        else:
            # Do literal exact matching of filename to database.
            c = self.conn.execute(
                'select name, obs_id, detset, sample_start, sample_stop '
                'from files where name=?', (filename, ))
            rows = c.fetchall()

        if len(rows) == 0:
            if fail_ok:
                return None
            raise RuntimeError('No match found for "%s"' % filename)
        if len(rows) > 1:
            raise RuntimeError('Multiple matches found for "%s"' % filename)
        _, obs_id, detset, start, stop = tuple(rows[0])

        return {'obs_id': obs_id,
                'detsets': [detset],
                'sample_range': (start, stop)}

    def verify(self, prefix=None):

        """Check the filesystem for the presence of files described in the
        database.  Returns a dictionary containing this information in
        various forms; see code for details.

        This function is used internally by the drop_incomplete()
        function, and may also be useful for debugging file-finding
        problems.

        """
        if prefix is None:
            prefix = self.prefix

        # Check for the presence of each listed file.
        c = self.conn.execute('select name, obs_id, detset, sample_start '
                              'from files')
        rows = []
        for r in c:
            fp = os.path.join(prefix, r[0])
            rows.append((os.path.exists(fp), fp) + tuple(r))

        obs = OrderedDict()
        for r in rows:
            present, fullpath, name, obs_id, detset, sample_start = r
            if obs_id not in obs:
                obs[obs_id] = {'present': [],
                               'absent': []}
            if present:
                obs[obs_id]['present'].append((detset, sample_start))
            else:
                obs[obs_id]['absent'].append((detset, sample_start))

        # Make a detset, sample_start grid for each observation.
        grids = OrderedDict()
        for k, v in obs.items():
            items = v['present']
            detsets = list(set([a for a, b in items]))
            sample_starts = list(set([b for a, b in items]))
            grid = np.zeros((len(detsets), len(sample_starts)), bool)
            for a, b in items:
                grid[detsets.index(a), sample_starts.index(b)] = True
            grids[k] = {'detset': detsets,
                        'sample_start': sample_starts,
                        'grid': grid}

        return {'raw': rows,
                'obs_id': obs,
                'grids': grids}

    def drop_obs(self, obs_id):
        """Delete the specified obs_id from the database.  Returns a list of
        files that are no longer covered by the database (with prefix).

        """
        # What files does this affect?
        c = self.conn.execute('select name from files where obs_id=?',
                              (obs_id,))
        affected_files = [os.path.join(self.prefix, r[0]) for r in c]
        # Drop them.
        self.conn.execute('delete from frame_offsets where file_name in '
                          '(select name from files where obs_id=?)',
                          (obs_id,))
        self.conn.execute('delete from files where obs_id=?',
                          (obs_id,))
        self.conn.commit()
        return affected_files

    def drop_detset(self, detset):
        """Delete the specified detset from the database.  Returns a list of
        files that are no longer covered by the database (with
        prefix).

        """
        # What files does this affect?
        c = self.conn.execute('select name from files where detset=?',
                              (detset,))
        affected_files = [os.path.join(self.prefix, r[0]) for r in c]
        # Drop them.
        self.conn.execute('delete from frame_offsets where file_name in '
                          '(select name from files where detset=?)',
                          (detset,))
        self.conn.execute('delete from files where detset=?',
                          (detset,))
        self.conn.commit()
        return affected_files

    def drop_incomplete(self):
        """Compare the files actually present on the system to the ones listed
        in this database.  Drop detsets from each observation, as
        necessary, such that the database is consistent with the file
        system.

        Returns a list of files that are on the system but are no
        longer included in the database.

        """
        affected_files = []
        scan = self.verify()
        for obs_id, info in scan['grids'].items():
            # Drop any detset that does not have complete sample
            # coverage.
            detset_to_drop = np.any(~info['grid'], axis=1)
            for i in detset_to_drop.nonzero()[0]:
                affected_files.extend(
                    [r[0] for r in self.conn.execute(
                        'select name from files where obs_id=? and detset=?',
                        (obs_id, info['detset'][i]))])
                self.conn.execute(
                    'delete from files where obs_id=? and detset=?',
                    (obs_id, info['detset'][i]))
        # Drop any detset that no longer appear in any files.
        self.conn.execute('delete from detsets where name not in '
                            '(select distinct detset from files)')
        self.conn.commit()
        self.conn.execute('vacuum')

        # Return the full paths of only the existing files that have
        # been dropped from the Db.
        path_map = {r[2]: r[1] for r in scan['raw'] if r[0]}
        return [r[1] for r in scan['raw'] if r[2] in affected_files]

    def get_file_list(self, fout=None):
        """Returns a list of all files in the database, without the file
        prefix, sorted by observation / detset / sample_start.  This
        is the sort of list one might use with rsync --files-from.

        If you pass an open file or filename to fout, the names will
        be written there, too.

        """
        c = self.conn.execute('select name from files order by '
                              'obs_id, detset, sample_start')
        output = [r[0] for r in c]
        if fout is not None:
            if isinstance(fout, str):
                assert(not os.path.exists(fout))
                fout = open(fout, 'w')
            for line in output:
                fout.write(line+'\n')
        return output


# Support functions for diff / patch

def _reconcile_gen(c0, c1, order_keys=None):
    """Generator used to compare rows in cursor c0 to those in cursor
    c1. The cursors are assumed to return sorted results, and that the
    sorting matches the sorting of tuple(row) (if order_keys is None)
    or tuple(row[k] or k in order_keys).

    Yields results of the form (mtype, item) where mtype is one of:
    - 'left-only': the item was found in c0 only
    - 'right-only': the item was found in c1 only
    - 'both': the item was found in both c0 and c1.

    """
    def _cmpbl(row):
        if row is None:
            return None
        if order_keys is None:
            return tuple(row)
        return tuple(row[k] for k in order_keys)

    r0, r1 = None, None
    while True:
        if r0 is None:
            r0 = c0.fetchone()
            t0 = _cmpbl(r0)
        if r1 is None:
            r1 = c1.fetchone()
            t1 = _cmpbl(r1)
        if r0 is None and r1 is None:
            break
        if r1 is None or (r0 is not None and t0 < t1):
            yield ('left-only', r0)
            r0 = None
        elif r0 is None or (r1 is not None and t1 < t0):
            yield ('right-only', r1)
            r1 = None
        else:
            assert t0 == t1
            yield('both', r0)
            r0 = r1 = None


def _count_and_collect(c0, c1, collect=[], order_keys=None):
    """Compare rows from cursors c0 and c1 using _reconcile_gen.  Count
    how many rows are in classes "left-only", "right-only", or "both".
    If any of those classes is listed in collect, then all rows
    matching that class are gathered into a ResultSet and returned.

    Returns:
      counts: dict of number of events.
      collections: dict with any ResultSet objects requested through
        collect kw.

    """
    keys = [col[0] for col in c1.description]
    collections = {k: ResultSet(keys=keys)
                   for k in collect}
    counts = {}
    for mtype, item in _reconcile_gen(c0, c1, order_keys=order_keys):
        counts[mtype] = counts.get(mtype, 0) + 1
        if mtype in collections:
            collections[mtype].rows.append(tuple(item))
    return counts, collections


def diff_obsfiledbs(db_left, db_right, return_detail=False):
    """Examine all records in two obsfiledbs and construct a list of
    changes that could made to db_left in order to make it match
    db_right.

    Returns a dict with following entries:

    - ``different`` (bool): whether the two databases carry different
      information.
    - ``patchable`` (bool): whether the function was able to construct
      patching instructions.
    - ``unpatchable_reason`` (str): if not patchable, a string
      explaining why.
    - ``detail`` (various): if not patchable, and return_detail, then
      this will contain detail about the offending data (e.g. obs rows
      in the two dbs that contain discrepant data).
    - ``patch_data`` (dict): if patchable, the data needed to patch
      db_left.  The fields are:

      - ``remove_files`` (list): entries to remove from obs table.
      - ``remove_dets`` (list): entries to remove from tags table.
      - ``new_files`` (ResultSet): new data for obs table -- iteration
        will yield dict that can be passed directly to db.update_obs.
      - ``new_dets`` (ResultSet): new data for dets table.


    Notes:

    In the present implementation, only changes involving adding rows
    to db_left (either whole obs rows or tag rows) will yield a
    patchable result.  Cases where some data has changed, or rows
    deleted, will simply return as unpatchable.

    """
    if isinstance(db_left, str):
        db_left = ObsFileDb.from_file(db_left, force_new_db=False)
    if isinstance(db_right, str):
        obsdb_right = ObsFileDb.from_file(db_right, force_new_db=False)

    def failure_declaration(reason, detail=None):
        if not return_detail:
            detail = None
        return {'different': True,
                'patchable': False,
                'unpatchable_reason': reason,
                'detail': detail}

    # Check meta.
    cursors = [
        db.conn.execute(
            'select * from meta order by param') for db in [db_left, db_right]]
    counts, collections = _count_and_collect(cursors[0], cursors[1],
                                             collect=['left-only', 'right-only'])
    if counts.get('left-only') or counts.get('right-only'):
        return failure_declaration(
            'Databases have different meta params.',
            detail=collections)

    # Check frame_offsets are not populated.
    foffs = [ResultSet.from_cursor(db.conn.execute(
        'select * from frame_offsets')) for db in [db_left, db_right]]
    if not (len(foffs[0]) == len(foffs[1]) == 0):
        return failure_declaration(
            'One or other database has non-empty frame_offsets table.',
            detail=metas)

    # The detsets table.
    cursors = [
        db.conn.execute(
            'select name, det from detsets order by name, det')
        for db in [db_left, db_right]]
    counts, collections = _count_and_collect(cursors[0], cursors[1], collect=['right-only'])
    bad_count = counts.get('left-only', 0)
    if bad_count > 0:
        return failure_declaration(
            f'db_left contains {bad_count} dets entries not found in db_right.',
            detail=collections.get('left-only'))
    new_dets = collections['right-only']

    # The files table.
    cursors = [
        db.conn.execute(
            'select * from files order by obs_id, detset, name, sample_start')
        for db in [db_left, db_right]]
    counts, collections = _count_and_collect(
        cursors[0], cursors[1], collect=['right-only'],
        order_keys=['obs_id', 'detset', 'name', 'sample_start'])
    bad_count = counts.get('left-only', 0)
    if bad_count > 0:
        return failure_declaration(
            f'db_left contains {bad_count} files entries not found in db_right.',
            detail=collections.get('left-only'))
    new_files = collections['right-only']
    new_files.keys[new_files.keys.index('name')] = 'filename'

    # Ok finally
    pd = {
        'remove_files': [],
        'remove_dets': [],
        'new_files': new_files,
        'new_dets': new_dets,
    }
    different = len(new_files) > 0 or len(new_dets) > 0
    return {
        'different': different,
        'patchable': True,
        'patch_data': pd,
    }


def patch_obsfiledb(patch_data, target_db):
    """Update an ObsFileDb with a batch of changes.

    Args:
      target_db (ObsFileDb): the database where changes should be made.
      patch_data (dict): patch information, as returned by
        diff_obsfiledbs.

    """
    assert len(patch_data['remove_files']) == 0
    assert len(patch_data['remove_dets']) == 0

    for file_entry in patch_data['new_files']:
        target_db.add_obsfile(**file_entry, commit=False)

    for det_entry in patch_data['new_dets']:
        target_db.add_detset(det_entry['name'], [det_entry['det']], commit=False)

    target_db.conn.commit()


def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser(
            epilog="""For details of individual modes, pass a dummy database argument
            followed by the mode and -h, e.g.: "%(prog)s x files -h" """)

    parser.add_argument('filename', help="Path to an ObsFileDb.",
                        metavar='obsfiledb.sqlite')

    cmdsubp = parser.add_subparsers(
        dest='mode')

    # "files"
    p = cmdsubp.add_parser(
        'files', usage="""Syntax:

    %(prog)s
    %(prog)s --all
    %(prog)s --clean

        This will print out a list of the files in the db,
        along with obs_id and detset.  Only a few lines
        will be shown, unless --all is passed.  To get a
        simple list of all files (for rsync or something),
        pass --clean.
        """,
        help="List the files referenced in the database.")

    p.add_argument('--clean', action='store_true',
                   help="Print a simple list of all files (for script digestion).")
    p.add_argument('--all', action='store_true',
                   help="Print all files, not an abbreviated list.")

    # "reroot"
    p = cmdsubp.add_parser(
        'reroot', help=
        "Batch change filenames (by prefix) in the database.",
        usage="""Syntax:

    %(prog)s old_prefix new_prefix [output options]

Examples:

    %(prog)s /path/on/system1 /path/on/system2 -o my_new_manifest.sqlite
    %(prog)s /path/on/system1 /new_path/on/system1 --overwrite
    %(prog)s ./result1/obs_12345.h5 ./result2/obs_12345.h5 --overwrite

        These operations will create a duplicate of the source
        ObsFileDb, with only the filenames (potentially) altered.  Any
        filename that starts with the first argument will be changed,
        in the output, to instead start with the second argument.
        When you do this you must either say where to write the output
        (-o) or give the program permission to overwrite your input
        database file.  Note that the first argument need not match
        all entries in the database; you can use it to pick out a
        subset (even a single entry).  """)
    p.add_argument('old_prefix', help=
                   "Prefix to match in current database.")
    p.add_argument('new_prefix', help=
                   "Prefix to replace it with.")
    p.add_argument('--overwrite', action='store_true', help=
                   "Store modified database in the same file.")
    p.add_argument('--output-db', '-o', help=
                   "Store modified database in this file.")
    p.add_argument('--dry-run', action='store_true', help=
                   "Run the conversion steps but do not write the results anywhere.")

    # "diff"
    p = cmdsubp.add_parser(
        'diff', help=
        "Check database against some upstream target; report diff/patchability.",
        usage="""Syntax:

    %(prog)s [output options]
        """)
    p.add_argument('upstream_db')
    p.add_argument('--patch', action='store_true', help=
                   "If possible, patch target to match the upstream.")

    # "fix-db"
    p = cmdsubp.add_parser(
        'fix-db', help=
        "Upgrade database (schema fixes, etc).",
        usage="""Syntax:

    %(prog)s [output options]
        """)
    p.add_argument('--overwrite', action='store_true', help=
                   "Store modified database in the same file.")
    p.add_argument('--output-db', '-o', help=
                   "Store modified database in this file.")
    p.add_argument('--dry-run', action='store_true', help=
                   "Run the conversion steps but do not write the results anywhere.")

    return parser

def main(args=None, parser=None):
    """Entry point for the so-metadata tool."""
    if args is None:
        args = sys.argv[1:]
    elif not isinstance(args, argparse.Namespace):
        parser = get_parser()
        args = parser.parse_args(args)

    if args.mode is None:
        args.mode = 'summary'

    db = ObsFileDb.from_file(args.filename, force_new_db=False)

    if args.mode == 'files':
        # Get all files.
        rows = db.conn.execute(
            'select obs_id, name, detset from files '
            'order by obs_id, detset, name').fetchall()

        if args.clean:
            for obs_id, filename, detset in rows:
                print(filename)
        else:
            fmt = '  {obs_id} {detset} {filename}'
            hdr = fmt.format(obs_id="obs_id", detset="detset", filename="Filename")
            print(hdr)
            print('-' * (len(hdr) + 20))
            n = len(rows)
            if n > 20 and not args.all:
                rows = rows[:10]
            for obs_id, filename, detset in rows:
                print(fmt.format(obs_id=obs_id, filename=filename, detset=detset))
            if len(rows) < n:
                print(fmt.format(obs_id='...', filename='+%i others' % (n - len(rows)), detset=''))
                print('(Pass --all to show all results.)')
            print()

    elif args.mode == 'reroot':
        # Reconnect with write?
        if args.overwrite:
            if args.output_db:
                parser.error("Specify only one of --overwrite or --output-db.")
            db = ObsFileDb.from_file(args.filename, force_new_db=True)
            args.output_db = args.filename
        else:
            if args.output_db is None:
                parser.error("Specify an output database name with --output-db, "
                             "or pass --overwrite to clobber.")
            db = ObsFileDb.from_file(args.filename, force_new_db=True)

        # Get all files matching this prefix ...
        c = db.conn.execute('select name from files '
                            'where name like "%s%%"' % (args.old_prefix))
        rows = c.fetchall()
        print('Found %i records matching prefix ...'
               % len(rows))

        print('Converting to new prefix ...')
        n_examples = 1

        if not args.dry_run:
            c = db.conn.cursor()

        for (name, ) in rows:
            new_name = args.new_prefix + name[len(args.old_prefix):]
            if n_examples > 0:
                print(f'  Example: converting filename\n'
                      f'      "{name}"\n'
                      f'    to\n'
                      f'      "{new_name}"')
                n_examples -= 1
            if not args.dry_run:
                c.execute('update files set name=? where name=?', (new_name, name))

        print('Saving to %s' % args.output_db)
        if not args.dry_run:
            db.conn.commit()
            c.execute('vacuum')
            db.to_file(args.output_db)

    elif args.mode == 'diff':
        print(f'Comparing to {args.upstream_db} ...')
        db = ObsFileDb(args.filename)
        db_right = ObsFileDb(args.upstream_db)
        report = diff_obsfiledbs(db, db_right)
        if not report['different']:
            print(' ... databases are in sync.')
        elif report['patchable']:
            print(' ... upstream is different, but the target db can be patched to match.')
        else:
            print(' ... upstream and target have irreconcilable differences.')
            parser.exit(1)

        if args.patch and report['different']:
            print()
            print('Patching ...')
            patch_obsfiledb(report['patch_data'], db)
            print(' ... done')
            print()

    elif args.mode == 'fix-db':
        # Reconnect with write?
        if args.overwrite:
            if args.output_db:
                parser.error("Specify only one of --overwrite or --output-db.")
            db = ObsFileDb.from_file(args.filename, force_new_db=True)
            args.output_db = args.filename
        else:
            if args.output_db is None:
                parser.error("Specify an output database name with --output-db, "
                             "or pass --overwrite to clobber.")
            db = ObsFileDb.from_file(args.filename, force_new_db=True)

        # Get version ...
        v = db._get_version()
        print(f'Database reports as version = {v}')

        changes = False
        if v == 1:
            # Copy detsets to new table, where uniqueness has been
            # relaxed.  Re-do the meta table, where uniqueness is
            # enforced to prevent lots of redundant rows added by
            # _create().
            changes = True
            for line in [
                    'drop table meta',
                    'alter table detsets rename to old_detsets',
                    '*',
                    'insert into detsets (name, det) select name, det from old_detsets',
                    'drop table old_detsets',
            ]:
                if line == '*':
                    print('Creating updated tables.')
                    db._create()
                    continue
                print(f'Running: {line}')
                db.conn.execute(line)
            print()

            for index in TABLE_DEFS['_indices']:
                print('Adding indexes...')
                db.conn.execute(index)

        elif v == 2:
            changes = True
            for index in TABLE_DEFS['_indices']:
                print('Adding indexes...')
                db.conn.execute(index)
            print('Bumping version')
            db.conn.execute('insert or replace into meta (param,value) values (?,?)',
                            ('obsfiledb_version', 3))

        if changes:
            print('Saving to %s' % args.output_db)
            if not args.dry_run:
                db.conn.commit()
                db.conn.execute('vacuum')
                db.to_file(args.output_db)
        else:
            print('No changes to make.')

    else:
        parser.error(f'Unimplemented mode, "{args.mode}".')



# Copyright (c) 2023-2025 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simulated optical pickup from various sources.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_simulate_scan_synchronous_signal(operators):
    """Add commandline args and operators for scan-synchronous signal.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.SimScanSynchronousSignal(name="sim_sss", enabled=False))


@workflow_timer
def simulate_scan_synchronous_signal(job, otherargs, runargs, data):
    """Simulate scan-synchronous signal.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if otherargs.realization is not None:
        job_ops.sim_sss.realization = otherargs.realization

    if job_ops.sim_sss.enabled:
        job_ops.sim_sss.detector_pointing = job_ops.det_pointing_azel_sim
        job_ops.sim_sss.stokes_weights = job_ops.weights_azel
        job_ops.sim_sss.apply(data)


def setup_simulate_hwpss_signal(operators):
    """Add commandline args and operators for HWP synchronous signal.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.SimHWPSS(name="sim_hwpss", enabled=False))


@workflow_timer
def simulate_hwpss_signal(job, otherargs, runargs, data):
    """Simulate HWP synchronous signal.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.sim_hwpss.enabled and job_ops.sim_ground.hwp_angle is not None:
        job_ops.sim_hwpss.hwp_angle = job_ops.sim_ground.hwp_angle
        job_ops.sim_hwpss.detector_pointing = job_ops.det_pointing_azel
        job_ops.sim_hwpss.stokes_weights = job_ops.weights_azel
        job_ops.sim_hwpss.apply(data)



# Copyright (c) 2025-2025 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import numpy as np

from astropy import units as u

import traitlets

import toast
from toast.timing import function_timer, Timer
from toast.traits import (
    trait_docs,
    Int,
    Unicode,
    Float,
    Quantity,
)
from toast.ops.operator import Operator
from toast.utils import Logger
from toast.observation import default_values as defaults


@trait_docs
class DetTrendCuts(Operator):
    """Use piecewise changes in slope to detect unlocked detectors.

    """

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    max_trend = Float(
        2.5,
        help="Slope (in phase units) at which detectors are unlocked"
    )

    t_piece = Quantity(
        10 * u.second,
        help="Time for the piecewise chunks",
    )

    max_samples = Int(
        500,
        help="Max samples per chunk for trend calc.  Longer chunks are decimated",
    )

    times = Unicode(defaults.times, help="Observation shared key for timestamps")

    det_data = Unicode(
        defaults.det_data,
        help="Observation detdata key for signal",
    )

    det_mask = Int(
        defaults.det_mask_invalid,
        help="Bit mask value for per-detector flagging",
    )

    shared_flags = Unicode(
        defaults.shared_flags,
        allow_none=True,
        help="Observation shared key for common flags",
    )

    shared_flag_mask = Int(
        defaults.shared_mask_invalid,
        help="Bit mask value for optional shared flagging",
    )

    det_flags = Unicode(
        defaults.det_flags,
        allow_none=True,
        help="Observation detdata key for detector flags",
    )

    det_flag_mask = Int(
        defaults.det_mask_invalid,
        help="Bit mask value for detector sample flagging",
    )

    trend_mask = Int(
        defaults.det_mask_invalid,
        help="Bit mask to apply to detectors that are outlier trends",
    )

    @traitlets.validate("det_mask")
    def _check_det_mask(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("Det mask should be a positive integer")
        return check

    @traitlets.validate("trend_mask")
    def _check_trend_mask(self, proposal):
        check = proposal["value"]
        if check < 0:
            raise traitlets.TraitError("Trend mask should be a positive integer")
        return check

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()

        for ob in data.obs:
            # Relative timestamps within the observation
            reltime = ob.shared[self.times].data.copy()
            t_0 = reltime[0]
            reltime -= t_0

            # Compute chunking
            slices = self._compute_chunk_slices(ob, reltime)

            # Flag detectors with any chunk-wise slope beyond the limit
            trend_flags = dict()
            for det in ob.select_local_detectors(detectors, flagmask=self.det_mask):
                slopes = self._compute_trend(ob, reltime, slices, det)
                max_abs_slope = np.amax(np.absolute(slopes))
                if max_abs_slope > self.max_trend:
                    msg = f"{ob.name}:{det} max trend slope {max_abs_slope} is "
                    msg += f"greater than limit ({self.max_trend}), cutting."
                    log.debug(msg)
                    trend_flags[det] = self.trend_mask
                else:
                    msg = f"{ob.name}:{det} max trend slope {max_abs_slope} is "
                    msg += f"less than limit ({self.max_trend}), keeping."
                    log.verbose(msg)

            # Update per-detector flags
            ob.update_local_detector_flags(trend_flags)

    def _compute_chunk_slices(self, obs, reltime):
        # Sample rate
        (rate, dt, dt_min, dt_max, dt_std) = toast.utils.rate_from_times(reltime)

        # Approximate target samples per chunk
        n_samples_per_piece = int(self.t_piece.to_value(u.second) * rate)

        # Distribute total samples into uniform pieces
        dist = toast.dist.distribute_uniform(len(reltime), n_samples_per_piece)

        # For each chunk we either use all samples or we downsample to meet the
        # max samples per chunk.
        slices = list()
        for offset, n_elem in dist:
            if n_elem > self.max_samples:
                step = 1 + n_elem // self.max_samples
            else:
                step = 1
            slices.append(slice(offset, offset + n_elem, step))
        return slices

    def _compute_trend(self, obs, reltime, slices, det):
        # This is a copy of the algorithm that is used in 
        # sotodlib.tod_ops.flags.get_trending_flags()
        signal = obs.detdata[self.det_data][det]
        flags = obs.shared[self.shared_flags].data & self.shared_flag_mask
        if self.det_flags is not None:
            flags |= obs.detdata[self.det_flags][det] & self.det_flag_mask
        slopes = list()
        for slc in slices:
            good = flags[slc] == 0
            t = reltime[slc][good]
            s = signal[slc][good]

            t_mean = np.mean(t)
            t_var = np.var(t, mean=t_mean)
            s_mean = np.mean(s)
            ts_mean = np.mean(t * s)

            slopes.append((ts_mean - t_mean * s_mean) / t_var)
        return np.array(slopes)

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "meta": list(),
            "shared": [self.times],
            "detdata": [self.det_data],
            "intervals": list(),
        }
        if self.shared_flags is not None:
            req["shared"].append(self.shared_flags)
        if self.det_flags is not None:
            req["detdata"].append(self.det_flags)
        return req

    def _provides(self):
        return dict()



try:
    import socs.agents.smurf_file_emulator.agent as sfe
except ModuleNotFoundError:
    sfe = None
import sotodlib.io.bookbinder as bb
import numpy as np
import so3g
from spt3g import core
import os
from unittest import mock
import unittest
import tempfile

def load_data(files, data_name='data'):
    """
    Load in G3 data and timestamps
    """
    it = bb.get_frame_iter(files)
    data = []
    times = []
    for frame in it:
        if frame.type != core.G3FrameType.Scan:
            continue
        data.append(frame[data_name].data)
        times.append(np.array(frame[data_name].times) / core.G3Units.s)
    data = np.hstack(data)
    times = np.hstack(times)
    return times, data

class BookbinderTest(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory()
        print("tempdir: {}".format(self.tempdir.name))
        self.l2_dir = os.path.join(self.tempdir.name, 'l2_archive')
        self.l3_dir = os.path.join(self.tempdir.name, 'books')
        os.makedirs(self.l2_dir, exist_ok=True)
        os.makedirs(self.l3_dir, exist_ok=True)

    def tearDown(self):
        self.tempdir.cleanup()

    def write_data(self, stream_id, duration, sample_rate, nchans,
                   drop_chance=0.1, start_offset=0):
        """
        Writes realistic smurf data using the smurf file emulator
        """
        frame_len = 5
        file_duration = 60*10
        mock_agent = mock.MagicMock()
        os.makedirs(self.l2_dir, exist_ok=True)

        parser = sfe.make_parser()
        args = parser.parse_args(args=[
            '--stream-id', stream_id, 
            '--base-dir', self.l2_dir,
            '--file-duration', str(file_duration),
            '--frame-len', str(frame_len),
            '--sample-rate', str(sample_rate),
            '--drop-chance', str(drop_chance),
            '--nchans', str(nchans), ])
        em = sfe.SmurfFileEmulator(mock_agent, args)
        session = mock.MagicMock()
        session.data = {}
        em.uxm_setup(session, {'sleep': False})
        em.stream(session, dict(
            duration=duration, use_stream_between=True, start_offset=start_offset
        ))
        return session.data['g3_files']
    
    def test_smurf_stream_processor(self):
        if sfe is None:
            print(
                "No socs module. "
                "Skipping BookbinderTest.test_smurf_stream_processor"
            )
            return
        l2_files = self.write_data('test', 60, 200, 5, drop_chance=0.1)

        in_ts, in_data = load_data(l2_files, data_name='data')

        ts, mask = bb.fill_time_gaps(in_ts)
        frame_idxs = np.zeros_like(ts, dtype=np.int32)
        file_idxs = np.zeros_like(np.unique(frame_idxs), dtype=np.int32)

        obsid = 'obsid'
        bookid = 'bookid'
        readout_ids = [str(i) for i in range(len(in_data))]
        ssp = bb.SmurfStreamProcessor(
            obsid, l2_files, bookid, readout_ids, allow_bad_timing=True,
        )
        ssp.preprocess()
        ssp.bind(self.l3_dir, ts, frame_idxs, file_idxs)

        out_ts, out_data = load_data(ssp.out_files, data_name='signal')

        self.assertTrue(np.all(out_ts[mask] == in_ts))
        self.assertTrue(np.all(out_data[:, mask] == in_data))

if __name__ == '__main__':
    unittest.main()



import unittest

from sotodlib.core import metadata

from ._helpers import create_outdir, mpi_multi


@unittest.skipIf(mpi_multi(), "Running with multiple MPI processes")
class TestProdDb(unittest.TestCase):
    def setUp(self):
        # Init.
        scheme = metadata.ManifestScheme()
        scheme.add_exact_match('array')
        scheme.add_range_match('time', dtype='float')
        scheme.add_data_field('also_data')
        self.scheme = scheme
        self.manifest = metadata.ManifestDb(scheme=scheme)

    def test_basic(self):
        manifest = self.manifest
        # Check file creation.
        file_id = manifest._get_file_id('test')
        assert(file_id is None) # New file_id should not be created.
        file_id = manifest._get_file_id('test', True)
        assert(file_id is not None) # New file_id should be created.
        assert(file_id == manifest._get_file_id('test')) # This file should now exist.

        # Add an indexed entry, with a new file attached.
        manifest.add_entry({'array': 'pa3', 'time': (12000., 13000.),
                            'also_data': 'wafered'},
                           'test2',create=True)

        for bad_time in [11999, 13000]:
            assert manifest.match({'array': 'pa3', 'time': bad_time}) is None
        for good_time in [12000, 12123]:
            assert manifest.match({'array': 'pa3',
                                   'time': good_time})['filename'] == 'test2'
        assert manifest.match({'array': 'pa2',
                               'time': 12000}) is None # Array does not match.

    def test_schema(self):
        print('\nCONSTRUCTED   :', self.scheme.cols)
        print('\nRECONSTRUCTED :', self.manifest.scheme.cols)
        assert (self.manifest.scheme.cols == self.scheme.cols)


if __name__ == '__main__':
    unittest.main()

